# -*- coding: utf-8 -*-
"""
Created on Fri Feb 14 11:50:29 2014

@author: Tobias Jachowski
"""
import collections
import getpass
import os
import shutil
import tempfile
import transaction
from functools import wraps, update_wrapper
from zc.lockfile import LockError
from ZODB import FileStorage, DB
from ZODB.POSException import ConnectionStateError

from . import config as cf
from . import helpers as hp
from . import update
from . import version
from .calibration import calibration as cb
from .calibration import Calibration, CalibrationSource
from .graph import GraphMember
from .gui import GRS
from .modification import Modification
from .picklable import Attributes
from .region import record as rc
from .region import Region, View, Record
from pyoti.plugins.modifications.generic import GenericMod


def if_open(func):
    @wraps(func)
    def func_wrapper(self, *args, **kwargs):
        if self.is_open:
            return func(self, *args, **kwargs)
        print("Experiment is closed!")
    return func_wrapper


def if_closed_return(value_or_func=None):
    def if_closed_return_decorator(func):
        @wraps(func)
        def func_wrapper(self, *args, **kwargs):
            if self.is_open:
                return func(self, *args, **kwargs)
            print("Experiment is closed!")
            try:
                value = value_or_func()
            except:
                value = value_or_func
            return value
        return func_wrapper
    return if_closed_return_decorator


class Experiment(object):
    """
    Provide functions to analyse tweezers data and save the analysis in a file:

    * create records (i.e. reading tweezers data),
    * construct groups, consisting of views and/or modifications:
        * create views (i.e. select timespan of data),
        * create modifications (i.e. modify the data),
        * change the timespan of views graphically,
        * change the parameters of modifications graphically,
        * concatenate and remove views (i.e. concatenate the data),
    * assemble/stack different groups,
    * manage records, views, and modifications,
    * delete records, views, and modifications.
    * load/save the records, views, and modifications and the graph structure

    Attributes
    ----------
    opened_last : Experiment
        References the `Experiment`, which opened the last experiment file.
    created_last : Experiment
        References the last created `Experiment`.
    """

    opened_last = None
    created_last = None

    def __init__(self, filename=None, directory=None, **kwargs):
        """
        Parameters
        ----------
        filename : str, optional
            The name of the experiment database file that should be opened.
            If filename is omitted, no file is opened.
        directory : str, optional
            The directory where the experiment is stored (defaults to '.').

        Attributes
        ----------
        filename : str
        version : str
        records : self.record(name)
            Offers easy access to Records via dotted notation:
            self.records.name is the same as calling self.record('name')
        views : self.view(name)
            Offers Easy access to Views via dotted notation: self.views.name is
            the same as calling self.view('name').
        modifications : self.modification(name)
            Offers Easy access to Modifications via dotted notation:
            self.modifications.name is the same as calling
            self.modification('name')
        cached_region : list of Regions
            The Region that is cached. It is stored in status.
        last_added_region : Region
            The Region that was added last to the experiment. It is stored in
            status.
        last_adjusted_region : Region
            The region that was adjusted last. It is stored in status.
        """
        self._grs = GRS()

        class AttributeCaller(object):
            def __init__(self, attribute_function, callback_function):
                self.attribute_function = attribute_function
                self.callback_function = callback_function
                update_wrapper(self, callback_function)

            def __getattr__(self, name):
                return self.attribute_function(name=name)

            def __call__(self, name=None, group=None):
                return self.callback_function(name=name, group=group)

        # Convenient objects to access records, views, and modifications via
        # dot notation: self.record.alpha or self.view.used, ...
        self.region = AttributeCaller(self._region, self._region)
        self.record = AttributeCaller(self._record, self._record)
        self.view = AttributeCaller(self._view, self._view)
        self.modification = AttributeCaller(self._modification,
                                            self._modification)
        # Backwards compatibility for "plural" notation
        self.regions = AttributeCaller(self._region, self._regions)
        self.records = AttributeCaller(self._record, self._records)
        self.views = AttributeCaller(self._view, self._views)
        self.modifications = AttributeCaller(self._modification,
                                             self._modifications)

        # Initialise attributes
        self._filename = None

        self._storage = None
        self._db = None
        self._transaction_manager = transaction.TransactionManager()
        self._connection = None

        self._tempdir = None

        if filename is not None:
            self.open(filename, directory)

        type(self).created_last = self

    def open(self, filename=None, directory=None, verbose=True):
        """
        Open experiment database file.
        """
        if self.is_open:
            if verbose:
                print("The following experiment file is already opened:\n",
                      "  '%s'\n" % self.filename,
                      "Close it, before you open a new file.")
            return False

        # Get and set database filename
        filename, absdir, absfile = hp.file_and_dir(filename, directory)

        if absfile is None:
            # Create a tempfile
            self._tempdir = tempfile.mkdtemp(prefix='pyoti-',)
            self._filename = tempfile.mktemp(suffix='.fs', prefix='',
                                             dir=self._tempdir)
            if verbose:
                print("No filename given. Switch on temporary mode.")
        else:
            self._filename = absfile

        if verbose:
            # Print whether a new experiment file was created or an old one was
            # opened
            if os.path.isfile(self.filename):
                print("Open experiment file:\n",
                      "  '%s'" % self.filename)
            else:
                print("Create experiment file:\n",
                      "  '%s'" % self.filename)

        # Try to open/create the experiment file
        try:
            # create/open db and TransactionManager
            self._storage = FileStorage.FileStorage(self.filename)
            self._db = DB(self._storage)
            self._connection = self._db.open(
                transaction_manager=self._transaction_manager)
        except LockError as err:
            if verbose:
                print("Cannot open the experiment file!\n",
                      "The database is locked by another process or user.\n",
                      "The original error message is: '%s'" % (err))
            return False
        except FileNotFoundError as err:
            if verbose:
                print("Cannot open the experiment file!\n",
                      "The file cannot be created!\n",
                      "The original error message is: '%s'" % (err))
            return False
        except Exception:
            raise  # raise unknown errors

        # Initialise the database of a new experiment file
        self._init_db()

        # Update the temporary absdir of graphroot. Make the attribute
        # volatile, so that ZODB doesn't store it.
        filename, absdir, absfile = hp.file_and_dir(self.filename)
        self._graphroot._v_absdir = absdir

        # Update the database of an experiment file created with an old version
        update.update_db(self)

        # Save the current timestamp
        # now = datetime.datetime.now()
        # current_date = now.ctime()
        # current_date_iso = now.isoformat()
        # global _current
        type(self).opened_last = self
        return True

    @if_open
    def _init_db(self, autosave=True):
        # create status, graphroot and version if new experiment
        if ('version' not in self._dbroot
            and ('views' in self._dbroot
                 or 'modifications' in self._dbroot
                 or 'records' in self._dbroot)):
            # Saving the database before updating it, would persistently break
            # the tweezersdata objects.
            # self._dbroot['version'] = '0.0.1'
            # autosave = False
            self.close()
            raise TypeError(("The experiment file '%s' was created with the "
                             "tweezersdata package.\n"
                             "This version existed before the pyoti, "
                             "investigator, or tethered_bead package.\n"
                             "Please, load the experiment file with the "
                             "tweezersdata package.") % self.filename)
        if 'status' not in self._dbroot:
            self._dbroot['status'] = Attributes()
            self._status.cached_region = []
            self._status.last_added_region = None
            self._status.last_adjusted_view = None
            autosave and self._save(description="Add status")
        if 'graphroot' not in self._dbroot:
            self._dbroot['graphroot'] = GraphMember(max_parents=0,
                                                    name='graphroot',
                                                    group='graphroot')
            autosave and self._save(description="Add graphroot")
        if 'version' not in self._dbroot:
            # Keep track of sourcecode used to create the database
            self._dbroot['version'] = version()
            autosave and self._save(description="Add version")

    @if_closed_return(True)
    def close(self, verbose=True, discard_temp=True):
        """
        Close the file the Experiment is stored in.
        """
        # Check if database was opened in temporary mode. If so, abort, close,
        # and delete temporary files.
        if discard_temp and self._tempdir is not None:
            print("Temporary experiment file to be deleted:\n  '%s'"
                  % self.filename)
            self.abort()
            self._close(verbose=verbose)
            self._rmtemp()
            return True

        return self._close(verbose=verbose)

    @if_closed_return(True)
    def _close(self, verbose=True):
        try:
            self._connection.close()
            self._db.close()
            self._storage.close()
            self._connection = None
            self._transaction_manager = transaction.TransactionManager()
            self._db = None
            self._storage = None
            self._filename = None
        except ConnectionStateError:
            if verbose:
                print("Cannot close experiment file due to unsaved changes.\n",
                      "If you really want to close without saving, execute",
                      "abort() first.")
            return False
        except Exception:
            raise  # raise unknown errors
        else:
            if verbose:
                print("Experiment closed.")
        return True

    def _rmtemp(self):
        if self._tempdir is None:
            return
        print("Switching off temporary mode.")
        shutil.rmtree(self._tempdir, ignore_errors=True)
        self._tempdir = None

    @if_open
    def create_record(self, name='alpha', group=None, calibration=None,
                      cfgfile=None, auto_add=True, **kwargs):
        """
        Open existing or create a new `Record`.

        Parameters
        ----------
        name : str
            The name of the `Record`. If a `Record` with the given `name`
            already exists in this `Experiment`, no new `Record` is created and
            the existing `Record` is returned.
        group : str, optional
            The name of the group, Defaults to `name`.
        calibration : Calibration or CalibrationSource, optional
            The `calibration` is used to create the new `Record`. If you
            provide no calibration, a generic calibration will be automatically
            created.
        cfgfile : str, optional
            The configuration file describing the setup specific parameters.
            Also see method `region.record.create_record()`.
        auto_add : bool, optional
            If `auto_add` == True, the new `Record` is automatically added to
            this `Experiment` by calling `add_record()`.
        **kwargs
            Keyword arguments are passed to function
            `pyoti.region.record.create_record()`

        Returns
        -------
        Record

        Notes
        -----
        See attribute `records` and method `record()` for retrieving
        already added records.

        """
        # set default group to name
        group = group or name

        # Try to find an already created record
        record = self.record(name=name, group=group)

        # create a new record with name and group
        if record is None:
            # Create a standard calibration
            if calibration is None:
                calibration = create_calibration()

            if isinstance(calibration, CalibrationSource):
                calibration = Calibration(calibration)

            # create the record
            record = rc.create_record(calibration, name=name, group=group,
                                      cfgfile=cfgfile, root=self._graphroot,
                                      **kwargs)

        # record could not be created
        if record is None:
            print("Could not create record")
            return None

        # automatically add the record to this experiment (`self._graphroot`)
        if auto_add:
            record = self.add_record(record=record)

        return record

    @if_open
    def add_record(self, record=None, set_last_added_region=True):
        """
        Add a Record to self, if a Record with the same name does not exist.

        Parameters
        ----------
        record : Record
            The Record to be added to self. Do not add the `record`, if a
            Record with the same name is already saved in self. Instead, return
            the already saved Record.
        set_last_added_region : bool
            If True and `record` was added, set self.last_added_region to the
            added Record.

        Returns
        -------
        Record
            This is either the added or already exising Record. If `record`
            could not be added nor a Record with the same name already exists,
            return None.
        """
        # read in record
        # If record already exists, do nothing.

        if record is None or not isinstance(record, Record):
            print("Can only add instances of `Record`!")
            return None

        # Check if record with same name already exists
        record_ = self.record(name=record.name, group=record.group)

        if record_ is None:
            # record with that name does not exist yet -> add it to self
            print("Adding record '" + record.name + "':")
            # add record to root graph member
            self._graphroot.add_child(record, set_changed=False,
                                      bidirectional=False)
            if set_last_added_region:
                self.last_added_region = record

            # Set datasource._root to self._graphroot for foreign records
            if (hasattr(record.datasource, '_root')
               and record.datasource._root is None):
                record.datasource._root = self._graphroot
        else:
            # record with that name does exist already -> set record to the
            # existing one
            record = record_
            print("Using already existing record '%s' and existing "
                  "calibration:" % (record.name))

        print(("  '" + record.datasource.name + "'"))
        print(("  '" + record.calibration.calibsource.name + "'"))

        return record

    @if_open
    def append_group(self, name, group=None, group_type='selection',
                     adjust=True, cfgfile=None, **kwargs):
        """
        Append a group to the last added group of this experiment. For details,
        parameters, and returns see method `add_group()`.
        """
        parent_region = self.last_added_region
        self.add_group(name,
                       parent_region,
                       group=group,
                       group_type=group_type,
                       adjust=adjust,
                       cfgfile=cfgfile,
                       **kwargs)

    @if_open
    def add_group(self,
                  name,
                  parent_region,
                  group=None,
                  parent_group=None,
                  group_type='selection',
                  adjust=True,
                  cfgfile=None,
                  traces=None,
                  description=None,
                  modclass=None,
                  **kwargs):
        """
        Add a group which will retrieve its data from a given `parent`.
        The type of the group is determined by the parameter `group_type`.
        The different types of groups are defined in the configfile (`cfgfile`,
        defaults to "groups.cfg").

        A group can be either (a) created for solely selecting a timespan of
        the data retrieved by the parent or (b) additionally modifying the
        data. In the case (a) `add_group()` creates a View by calling the
        method `add_view()`. In the case (b)  `add_group()` creates a View and
        additionally a Modification, by calling the method
        `add_modification()`.


        Parameters
        ----------
        name : str
            The name of the members of the group. For Modifications and their
            Views based, the name will be changed to name + "_mod".
        parent_region : str or GraphMember
            Name of the parent region, a GraphMember with the name, or the
            parent Region itself.
        group : str, optional
            The name of the group, Defaults to `name`.
        parent_group : str, optional
            Name of the group of the `parent_region`. Defaults to
            None or `parent_region.group`.
        group_type : str
            The name of the type of group defined in the file `cfgfile`.
        adjust : bool
            If True, the user is presented a graphical selection dialog (GRS)
            to select the timespan of the new group (i.e. the View).
            See also method `adjust_view()` and `adjust_modification()`.
        cfgfile : str, filepath, optional
            The configfile where the different types of groups are defined.
            Values can be overwritten by parameters `traces`, `description`,
            and `modclass`
            Defaults to "groups.cfg"
        traces : str, int, list of str/int, or slice
            Select the traces the user is shown to adjust the timespan from.
            The selection of the traces does *not* influence the traces
            contained in the new group (i.e. View).
            Takes precedence over the settings of `cfgfile`. Defaults to all
            traces available in `parent_region`.
        description : str
            Short description of what the selection of the timespan is used
            for. Takes precedence over the settings of `cfgfile`.
        modclass : Modification
            If a class for a Modification is given, create a modifying group
            by calling the method `add_modifiation()`, else create a selecting
            group by calling the method `add_view()`. Defaults to None.
            Takes precedence over the settings of `cfgfile`.
        **kwargs : optional
            Additional Arguments that are passed either to the method
            `add_view()` or the method `add_modification()`.

        Returns
        -------
        View or Modification
            Depending on the type of group, return either the added `View` or
            the `Modification`.
        """
        cfgfile = cfgfile or 'groups.cfg'
        cfg = cf.read_cfg_file(cfgfile)

        description = description or cf.get_cfg_option(cfg, sec=group_type,
                                                       opt='description')
        traces = (traces or cf.get_cfg_list(cfg, sec=group_type, opt='traces')
                  or None)
        modclass = modclass or cf.get_cfg_class(cfg, sec=group_type,
                                    std_mod='.plugins.modifications.generic',
                                    cls_opt='modclass')

        if not modclass:
            return self.add_view(name, parent_region, group=group,
                                 parent_group=parent_group, adjust=adjust,
                                 traces=traces, description=description,
                                 **kwargs)
        else:
            return self.add_modification(name, group=group,
                                         modclass=modclass,
                                         parent_region=parent_region,
                                         parent_group=parent_group,
                                         adjust=adjust,
                                         traces=traces,
                                         description=description, **kwargs)

    @if_open
    def add_view(self, name, parent_region, group=None, parent_group=None,
                 adjust=True, traces=None, description=None, **kwargs):
        """
        Create and adjust a View with `parent_region` as a parent. The View
        will retrieve its data from the `parent_region`. The timespan is set to
        the same as the one from the parent. The new `View` is automatically
        added to this `Experiment`. If a `View` with the `name` already exists
        in this `Experiment`, do not create a new `View`, but instead retrieve
        and return the existing one. Let the user choose the timespan of the
        View, if adjust is True.

        Parameters
        ----------
        name : str
            The name of the View.
        parent_region : str or GraphMember
            Name of the parent region, a GraphMember with the name, or the
            parent Region itself.
        group : str, optional
            Name of the group, defaults to `name`.
        parent_group : str, optional
            Name of the group of the `parent_region`. Defaults to
            None or `parent_region.group`.
        adjust : bool
            If True, the user is presented a graphical selection dialog (GRS)
            to select the timespan of the new View.
            See also method `adjust_view()`.
        traces : str, int, list of str/int, or slice
            Select the traces the user is shown to adjust the timespan from.
            The selection of the traces does *not* influence the traces
            contained in the new View. Defaults to all traces available in the
            Region determined by `parent_region`.
        description : str
            Short description of what the selection of the timespan is used
            for. Takes precedence over the settings of `cfgfile`.
        **kwargs : optional
            Additional arguments to initialize a View. See `View' for
            details. Only use, if you really now what you ar doing.

        Returns
        -------
        View
        """
        group = group or name
        view = self._add_view(name, group, parent_region,
                              parent_group=parent_group, **kwargs)
        if adjust:
            self.adjust_view(view, group=group, traces=traces,
                             description=description)

        return view

    @if_open
    def _add_view(self, name, group, parent_region, parent_group=None,
                  set_last_added_region=True, **kwargs):
        """
        Create a View with `parent_region` as a parent. The View will retrieve
        its data from the `parent_region`. The timespan is set to the same as
        the one from the parent. The new `View` is automatically added to this
        `Experiment`. If a `View` with the `name` already exists in this
        `Experiment`, do not create a new `View`, but instead retrieve and
        return the existing one.

        Parameters
        ----------
        name : str
            The name of the View.
        group : str
            The name of the group the View should be part of.
        parent_region : str or GraphMember or collections.Iterable of these
            Name of the parent region, a GraphMember with the name, or the
            parent Region itself.
        parent_group : str, optional
            Name of the group of the parent_region. Defaults to None or
            `parent_region.group`.
        set_last_added_region : bool
            If True and a new View was added, set self.last_added_region to the
            added View.
        **kwargs : optional
            Additional arguments to initialize a View. See `View' for
            details. Only use, if you really now what you ar doing.

        Returns
        -------
        View
        """
        # try to get already existing view
        view = self.view(name=name, group=group)

        # view does not exist -> create new view
        if view is None:
            parent = self.region(name=parent_region, group=parent_group)
            if parent is not None:
                # switch on caching of this Region's parent
                # and switch off caching of previous Region's parent
                self.set_cached_region(parent)

                print(("Creating view '" + name + "'."))
                view = View(parent=parent, caching=False, name=name,
                            group=group, **kwargs)
                if set_last_added_region:
                    self.last_added_region = view
            else:
                print("Could not find parent %s" % (parent_region))
        else:
            print(("Using already existing view '" + name + "':"))

        return view

    @if_open
    def show_view(self, view, group=None, traces=None, description=None,
                  xlim=None):
        """
        Graphically inspect a View.

        Parameters
        ----------
        view : str or GraphMember
            Name of the View, a GraphMember with the name, or the View itself.
        group : str, optional
            The name of the group. Defaults to None or `view.group`.
        traces : str, int, list of str/int, or slice
            Select the traces the user is shown to adjust the timespan from.
            The selection of the traces does *not* influence the traces
            contained in the View. Defaults to all traces available in the
            View determined by `view`.
        description : str
            Short description of what the selection of the timespan is used
            for.
        xlim : (float, float), optional
            The timespan (s) plotted.
            Defaults to (View.tmin, View.tmax) of the View determined by
            `view`.
        """
        view = self.view(name=view, group=group)
        if view is not None:
            # if traces is None or (isinstance(traces, list)
            # and len(traces) == 0):
            traces = view.traces_available(traces)
            # tc.normalize(traces)
            # traces = traces or view.traces
            if description is None:
                description = view.name
            self._grs.init_ifig(view.timevector,
                                view.get_data(traces=traces, copy=False),
                                view.samplingrate,
                                traces,
                                xlim=xlim,
                                description=description,
                                select=False)

    @if_open
    def adjust_view(self, view, group=None, traces=None, description=None,
                    xlim=None, set_last_adjusted_view=True):
        """
        Graphically adjust the timespan of a View.

        Parameters
        ----------
        view : str or GraphMember
            Name of the View, a GraphMember with the name, or the View itself.
        group : str, optional
            The name of the group. Defaults to None or `view.group`.
        traces : str, int, list of str/int, or slice
            Select the traces the user is shown to adjust the timespan from.
            The selection of the traces does *not* influence the traces
            contained in the View. Defaults to all traces available in the
            View determined by `view`.
        description : str
            Short description of what the selection of the timespan is used
            for.
        xlim : (float, float), optional
            The timespan (s) plotted the user can choose the timespan from.
            Defaults to (View.parent.tmin, View.parent.tmax) of the View
            determined by `view`.
        set_last_adjusted_view : bool
            If True, set self.last_adjusted_view to the View determined by
            `view`.
        """
        view = self.view(name=view, group=group)
        if view is not None:
            # if traces is None or (isinstance(traces, list)
            # and len(traces) == 0):
            traces = view.traces_available(traces)
            # tc.normalize(traces)
            # traces = traces or view.traces
            if description is None:
                description = view.name
            self._grs.init_ifig(view.parent.timevector,
                                view.parent.get_data(traces=traces,
                                                     copy=False),
                                view.parent.samplingrate,
                                traces,
                                tmin=view.tmin,
                                tmax=view.tmax,
                                xlim=xlim,
                                onselect_cb=view.set_timespan,
                                description=description)

            if set_last_adjusted_view:
                self.last_adjusted_view = view

    @if_open
    def add_modification(self, name, group=None, modclass=None,
                         parent_region=None, parent_group=None,
                         view_apply=None, mod_name_append='_mod', adjust=True,
                         traces=None, description=None, **kwargs):
        """
        Create a `Modification` and associate it to a View the Modification is
        applied to (View applied) and a View whose data is used to calculate
        its parameters from (View based).
        The name of the View applied is `name`, or `apply_view`, if given. The
        name of the View based and the Modification is determined by
        concatenating `name` and `mod_name_append`.
        If either of the Views does not exist, automatically create them with
        `parent_region` as the parent Region (details see `add_view()`)
        If `adjust`=True, let the user graphically adjust the timespan of the
        View based on. Otherwise, leave the timespan untouched, if the View
        based already exists, or set the timespan to the same as the
        `parent_region` if the View based has to be created.

        Parameters
        ----------
        name : str
            The first part of the name of the Modification, and, if
            `apply_view` is not given, the name of the `View` the Modification
            should be applied to. If a View with `name` name does not exist,
            create a new one. See `add_view()` for details.
        group : str, optional
            The name of the group. Defaults to `name`.
        modclass : Modification
            The class of the Modification that should be used.
        parent_region : str or GraphMember, optional
            Name of the parent region, a GraphMember with the name, or the
            parent Region itself. `parent_region` will be used for the creation
            of the View applied as well as the View based. If it is not
            supplied, the parent(s) of the View applied are used. See
            `add_view()` for details.
        parent_group : str, optional
            Name of the group of the parent_region. Defaults to None or
            `parent_region.group`.
        view_apply : str or GraphMember, optional
            The name of the `View` the Modification should be applied to. If a
            View with name `name` does not exist, create a new one, with
            parent `parent_region`. See `add_view()` for details.
        mod_name_append : str
            The name of the Modification and the View based are determined by
            concatenating 'name' and 'mod_name_append'.
        adjust : bool
            If True, the user is presented a graphical selection dialog (GRS)
            to select the timespan of the View based. See also method
            `adjust_modification()`.
        traces : str, int, list of str/int, or slice
            Select the traces the user is shown to adjust the timespan from.
            The selection of the traces does *not* influence the traces
            contained in the View. Defaults to all traces available in the
            Region determined by `parent_region`.
        description : str
            Short description of what the selection of the timespan is used
            for.
        **kwargs
            Keyword arguments are passed on to initializing class `modclass`
        """
        # Create or get a view the modification is applied to
        if view_apply is None and parent_region is None:
            raise TypeError("add_modification() is missing keyword argument "
                            "`parent_region` or `view_apply`")

        group = group or name

        # Create a new view_apply, parent_region needs to be given
        if view_apply is None:
            view_apply = self._add_view(name, group, parent_region,
                                        parent_group=parent_group)

        # Set parent region(s) for the view_based of modification to be the
        # same as the ones from view_apply, view_apply needs to be given
        if parent_region is None:
            # Get existing view
            view_apply = self.view(name=view_apply, group=group)
            parent_region = view_apply.parent.parents

        # Create modification
        mod_name = name + mod_name_append

        # Set modclass to be defaulting to GenericMod
        modclass = modclass or GenericMod

        modification = self._add_modification(mod_name, group, modclass,
                                              view_apply, parent_region,
                                              parent_group=parent_group,
                                              **kwargs)
        if adjust:
            self.adjust_modification(modification, group=group, traces=traces,
                                     description=description)

        return modification

    @if_open
    def _add_modification(self, name, group, modclass, view_apply,
                          parent_region, parent_group=None, **kwargs):
        """
        Create a `Modification` and associate it to a View the Modification is
        applied to (View applied) and a View whose data is used to calculate
        its parameters from (View based).
        The name of the Modification and, if `parent_region` is given, the View
        based is `name`. The class used to create the Modification is
        `modclass`.
        If the Modification with the name `name` does exist, simply return the
        modification.

        Parameters
        ----------
        name : str
            The name of the Modification and the View based. If a Modification
            with name `name` does not exist, create one. If `parent_region` is
            given, use the name also for the View based, that is created, too,
            if a View with the same name does not exist yet.
        group : str
            Name of the group.
        modclass : Modification
            The class of the Modification that should be used.
        view_apply : str or GraphMember
            The name of the `View` applied.
        parent_region : str or GraphMember or Iterable of str/GraphMember
            The name of the parent Region(s) that is used as a parent for View
            based, if Modification and View based do not exist and have to be
            created.
        parent_group : str, optional
            Name of the group of the parent_region. Defaults to None or
            `parent_region.group`.
        **kwargs
            Keyword arguments are passed on to initializing of class
            `modclass`, when a new Modification has to be created.

        Returns
        -------
        Modification
            The created or found Modification.
        """
        modification = self.modification(name=name, group=group)

        if modification is None:
            # Create new modificiation
            view_apply = self.view(name=view_apply, group=group)
            parent_region = self.region(name=parent_region, group=parent_group)

            if view_apply is not None and parent_region is not None:
                # modification should  have a parent_region/view_based so that
                # it can be referenced by traversing the graph from top to
                # bottom

                # switch on caching of this Region's parent and switch off
                # caching of previous Region's parent
                self.set_cached_region(parent_region)

                view_based = self._add_view(name, group, parent_region,
                                            set_last_added_region=False)

                print(("Creating modification '" + name + "'."))
                # Instantiate a Modification subclass
                # e.g. 'Offset' would correspond to subclass Offset
                modification = modclass(view_based=view_based,
                                        view_apply=view_apply, name=name,
                                        group=group, **kwargs)
            else:
                print("Could not find/use view_apply %s and/or parent_region "
                      "%s." % (view_apply, parent_region))
        else:
            print(("Using already existing modification '" + name + "':"))

        # initialize the modification of the modification
        # if this would not be done directly after having added th
        # modification the modification would first be calculated, when trying
        # to get data from the View the modification is applied to.
        # modification.evaluate()

        return modification

    @if_open
    def adjust_modification(self, modification, group=None, traces=None,
                            description=None, xlim=None):
        """
        Adjust the timespan of a View based of a Modification.

        Parameters
        ----------
        modification : str or GraphMember
            Name of the Modfication whose View based should be adjusted.
        group : str, optional
            The name of the group. Defaults to None or `modification.group`.
        traces : str, int, list of str/int, or slice
            Select the traces the user is shown to adjust the timespan from.
            The selection of the traces does *not* influence the traces
            contained in the View. Defaults to all traces available in the
            View determined by `view`.
        description : str
            Short description of what the selection of the timespan is used
            for.
        xlim : (float, float), optional
            The timespan (s) plotted the user can choose the timespan from.
            Defaults to (View.parent.tmin, View.parent.tmax) of the View
            determined by `view`.
        """
        modification = self.modification(modification, group=group)

        self.adjust_view(modification.view_based, group=group, traces=traces,
                         description=description, xlim=xlim)

        try:
            # Interact with the modification, after view has been selected
            modification.interact()
        except Exception as exc:
            print("Something went wrong!\n",
                  "Can not interact with the modification!\n",
                  "The original error message is: '%s'" % (exc))

    @if_open
    def concatenate(self, name, first_region, second_region, group=None,
                    first_group=None, second_group=None, adjust=True,
                    traces=None, description=None):
        """
        Concatenate two Regions as parent Regions for a new View with name
        `name`. If the View with the name `name` already exists, append the
        `second_region` to the parent Regions of the View with the name `name`.

        Parameters
        ----------
        name : str
            Name of the View, which should have the data of the concatenated
            Regions.
        first_region : str or GraphMember
            Name of the Region, a GraphMember with the name, or the Region
            itself, whose data should be concatenated with second_region.
        second_region : str or GraphMember
            Name of the Region, a GraphMember with the name, or the Region
            itself, whose data should be concatenated with first_region.
        group : str, optional
            Name of the group for the new concatenated view. Defaults to
            `name`.
        first_group : str, optional
            Name of the group of the first_region. Defaults to None or
            `first_region.group`.
        second_group : str, optional
            Name of the group of the second_region. Defaults to None or
            `second_region.group`.
        adjust : bool
            If True, the user is presented a graphical selection dialog (GRS)
            to select the timespan of the new View.
            See also method `adjust_view()`.
        traces : str, int, list of str/int, or slice
            Select the traces the user is shown to adjust the timespan from.
            The selection of the traces does *not* influence the traces
            contained in the View. Defaults to all traces available in the
            View determined by `view`.
        description : str
            Short description of what the selection of the timespan is used
            for.

        Returns
        -------
        View
            The View which presents the concatenated data.
        """
        group = group or name
        view = self._add_view(name, group, first_region,
                              parent_group=first_group)
        self.append_to(view, second_region, group=group,
                       append_group=second_group)
        if adjust:
            self.adjust_view(view, group=group, traces=traces,
                             description=description)
        return view

    @if_open
    def append_to(self, name, append_region, group=None, append_group=None,
                  after=None, after_group=None, before=None,
                  before_group=None, update_indices=True):
        """
        Appends a Region as a parent to a View.

        Parameters
        ----------
        name : str
            Name of a View a new parent should be appended to.
        append_region : str or GraphMember
            Name of a Region, a GraphMember with the name, or the Region
            itself, which should be appended as a parent.
        group : str, optional
            Name of the group for the view the data should be appended to.
            Defaults to `name`.
        append_group : str, optional
            Name of the group of the append_region. Defaults to None or
            `append_region.group`.
        after : str or GraphMember
            Name of a Region, a GraphMember with the name, or the Region
            itself, after which the new Region should be appended to.
        after_group : str, optional
            Name of the group of the after region. Defaults to None or
            `after_region.group`.
        before : str or GraphMember
            Name of a Region, a GraphMember with the name, or the Region
            itself, before which the new Region should be appended to. If after
            was given and could be found, before is ignored.
        before_group : str, optional
            Name of the group of the before region. Defaults to None or
            `before_region.group`.
        update_indices : bool, optional
            Update the indices of the child Regions. Defaults to True.

        Returns
        -------
        bool
            True if the region was appended or is already a parent. Otherwise
            return False.
        """
        if group is None:
            if isinstance(name, GraphMember):
                group = name.group
            else:
                group = name
        view = self.view(name=name, group=group)
        append_region = self.region(name=append_region, group=append_group)
        if view is not None and append_region is not None:
            after = self.region(name=after, group=after_group)
            before = self.region(name=before, group=before_group)
            return view.parent.add_parent(append_region, after=after,
                                          before=before,
                                          update_indices=update_indices)
        return False

    @if_open
    def remove_record(self, record, group=None, dereference_children=True):
        """
        Remove a Record with the name `name`.

        Remove all references from record to its parent (self._graphroot) and
        vice versa.

        Be careful, if you remove a record, all children, which have the record
        as their only root, will be removed, too!

        Parameters
        ----------
        record : str or GraphMember
            Name of the Record, a GraphMember with the name, or the Record
            itself that should be removed.
        group : str, optional
            Name of the group of the record. Defaults to None or
            `record.group`.

        Returns
        -------
        Record
            The removed record.
        """
        # removing a record
        record = self.record(name=record, group=group)
        if record is not None:
            # remove record from root graph member
            self._graphroot.remove_child(record)
            # dereference children from record
            if dereference_children:
                for child in record.children:
                    record.remove_child(child)
            # Remove reference of datasource to self._graphroot
            if hasattr(record.datasource, '_root'):
                record.datasource._root = None
        return record

    @if_open
    def remove_view(self, view, group=None):
        """
        Remove a View or a View based and its associated Modification with the
        name `view`.

        Remove all references from view to its parents (view.parent.parents
        and view.modifications_apply) and its children (view.children and
        view.modifications_based) and vice versa.

        Parameters
        ----------
        view : str or GraphMember
            Name of the View, a GraphMember with the name, or the View itself
            that should be removed.
        group : str, optional
            Name of the group of the view. Defaults to None or `view.group`.

        Returns
        -------
        View
            The removed view.

        Notes
        -----
        The structure for a View could look like this:

        parent.parents
              |
             view             <-- region
              |
          children
        """
        view = self.view(name=view, group=group)
        if view is not None:
            # First, delete relations to parents (view.parent is a MultiRegion)
            for parent in view.parent.parents:
                parent.remove_child(view.parent)
            for mod_apply in view.modifications_apply:
                mod_apply.remove_child(view)
            # Second, delete relations to children
            for mod_based in view.modifications_based:
                if mod_based.view_apply is not None:
                    mod_based.view_apply.remove_parent(mod_based)
            for child in view.children:
                child.remove_parent(view)

        return view

    @if_open
    def remove_modification(self, modification, group=None):
        """
        Remove a Modification with the name `modification`.

        Remove all references from modification to its parent (View based)
        and its child (View apply) and vice versa.

        Parameters
        ----------
        modification : str or GraphMember
            Name of the Modification, a GraphMember with the name, or the
            Modification itself that should be removed.
        group : str, optional
            Name of the group of the modification. Defaults to None or
            `modification.group`.

        Returns
        -------
        Modification
            The removed Modification.

        Notes
        -----
        The structure for a Modification could look like this:

        parent.parents
              |
          view_based          <-- region
              |  <- mod_based
              |
         modification
              |
              |  <- mod_apply
          view_apply
              |
          children
        """
        modification = self.modification(name=modification, group=group)
        if modification is not None:
            # First, delete relation to parent (view.parent is a MultiRegion)
            modification.view_based.remove_child(modification)
            # Second, delete relation to child
            modification.view_apply.remove_parent(modification)

        return modification

    @if_open
    def remove(self, name, group=None):
        """
        Call and return the return value of the function `remove_view()`.
        If a view with the name could not be found, try to remove a record.

        Parameters
        ----------
        name : str or GraphMember
        group : str, optional
            Name of the group of the graphmember to be removed. Defaults to
            None or `name.group`.
        """
        return self.remove_view(view=name, group=group) \
            or self.remove_record(record=name, group=group)

    @if_open
    def remove_from(self, view, remove_region, group=None, remove_group=None,
                    update_indices=True):
        """
        Remove a parent Region of a (concatenated) view. If the region is the
        last in the view, the view and all its descendants are removed, too.

        Removes all references from view to its parent remove_region and vice
        versa.

        Parameters
        ----------
        view : str or GraphMember
            Name of the View, a GraphMember with the name, or the View itself
            that the parent Region should be removed from.
        remove_region : str or GraphMember
            Name of the Region, a GraphMember with the name, or the Region
            itself that should be removed.
        group : str, optional
            Name of the group of the view. Defaults to None or `view.group`.
        remove_group : str, optional
            Name of the group of the remove region. Defaults to None or
            `remove_region.group`.
        update_indices : bool, optional
            Update the indices of the child Regions. Defaults to True.

        Returns
        -------
        Region
            The removed region.
        """
        view = self.view(name=view, group=group)
        remove_region = self.region(name=remove_region, group=remove_group)

        if view is not None and remove_region is not None:
            view.parent.remove_parent(remove_region,
                                      update_indices=update_indices)

        return remove_region

    @if_open
    def replace_in(self, view, remove_region, with_region, group=None,
                   remove_group=None, with_group=None, update_indices=True):
        """
        Replace a parent Region of a view with another one.

        Parameters
        ----------
        view : str or GraphMember
            Name of the View, a GraphMember with the name, or the View itself
            that the parent Region should be replaced in.
        remove_region
            Name of the Region, a GraphMember with the name, or the Region
            itself that should be replaced.
        with_region
            Name of the Region, a GraphMember with the name, or the Region
            itself that should replace the `remove_region`.
        group : str, optional
            Name of the group of the view. Defaults to None or `view.group`.
        remove_group : str, optional
            Name of the group of the remove region. Defaults to None or
            `remove_region.group`.
        with_group : str, optional
            Name of the group of the with region. Defaults to None or
            `with_region.group`.
        update_indices : bool, optional
            Update the indices of the child Regions. Defaults to True.

        Returns
        -------
        Region
            The replaced region.
        """
        view = self.view(name=view, group=group)
        with_region = self.region(name=with_region, group=with_group)
        remove_region = self.region(name=remove_region, group=remove_group)
        if (view is not None
                and remove_region is not None
                and with_region is not None
                and remove_region is not with_region):
            self.append_to(view, with_region, after=remove_region,
                           update_indices=update_indices)
            self.remove_from(view, remove_region,
                             update_indices=update_indices)

        return remove_region

    @if_open
    def members(self, name=None, group=None, instance_class=None, dft=False,
                level=-1):
        """
        Yields the members of group `group` with name `name` and class
        `instance_class`.

        Parameters
        ----------
        name : `instance_class`, GraphMember, or str, optional
            The name of the group that should be searched for in this
            Experiment.
        group : str, optional
            The name of the group, Defaults to None or `name.group`.
        instance_class : class, optional
            The class of the instance that is searched for and returned.
        dft : bool, optional
        level : int, optional

        Yields
        ------
        GraphMember
        """
        if name is not None \
                and (group is None
                     or (isinstance(name, GraphMember)
                         and name.group is group)) \
                and (instance_class is None
                     or isinstance(name, instance_class)):
            yield name
        else:
            if isinstance(name, GraphMember):
                name = name.name
            # get the members
            members = self._graphroot.members(name=name, group=group,
                                              instance_class=instance_class,
                                              includeself=False, dft=dft,
                                              level=level)
            for member in members:
                yield member

    @if_open
    def member(self, name=None, group=None, instance_class=None, dft=False,
               level=-1):
        """
        Returns the last member of group `group` with name `name` and class
        `instance_class`.

        Parameters
        ----------
        name : instance_class, GraphMember, or str, optional
            The name of the group that should be searched for in this
            Experiment.
        group : str, optional
            The name of the group, Defaults to None or `name.group`.
        instance_class : class, optional
            The class of the instance that is searched for and returned.
        dft : bool, optional
        level : int, optional

        Returns
        -------
        member : GraphMember
        """
        # First get the members
        members = self.members(name=name, group=group,
                               instance_class=instance_class, dft=dft,
                               level=level)
        # Usually, a group consists of the following parent -> child
        # combinations:
        # 1. Record
        # 2. MultiRegion -> View
        # 3. Modification -> View
        # 4. MultiRegion -> View -> Modfication -> View
        #
        # To always get the View and not the MultiRegion, if a Region is asked
        # for, return the last member found member
        last = None
        for member in members:
            last = member
        return last

    @if_open
    def _regions(self, name=None, group=None):
        """
        Finds and yields all Regions of this `Experiment`.

        Yields
        ------
        Region
        """
        return self.members(name=name, group=group, instance_class=Region)

    @if_open
    def _records(self, name=None, group=None):
        """
        Finds and yields all Records of this `Experiment`.

        Yields
        ------
        Record
        """
        return self.members(name=name, group=group, instance_class=Record,
                            level=1)

    @if_open
    def _views(self, name=None, group=None):
        """
        Finds and yields all Views of this `Experiment`.

        Yields
        ------
        View
        """
        return self.members(name=name, group=group, instance_class=View)

    @if_open
    def _modifications(self, name=None, group=None):
        """
        Finds and yields all Modifications of this `Experiment`.

        Yields
        ------
        Region
        """
        return self.members(name=name, group=group,
                            instance_class=Modification)

    @if_open
    def _region(self, name=None, group=None):
        """
        If a Region with the same name as `name` is already existing in self,
        return the already existing one. Otherwise, if `name` is a Region
        itself, return `name`. If None of this is True, return None.

        Parameters
        ----------
        name : str or GraphMember

        Returns
        -------
        Region or None

        Notes
        -----
        The returned Region can be either a Record or a View.
        """
        if name is None and group is None:
            return None
        return self.member(name=name, group=group, instance_class=Region)

    @if_open
    def _record(self, name=None, group=None):
        """
        If a Record with the same name as `name` is already existing in self,
        return the already existing one. Otherwise, if `name` is a Record
        itself, return `name`. If None of this is True, return None.

        Parameters
        ----------
        name : str or GraphMember

        Returns
        -------
        Record or None
        """
        if name is None and group is None:
            return None
        return self.member(name=name, group=group, instance_class=Record,
                           level=1)

    @if_open
    def _view(self, name=None, group=None):
        """
        If a View with the same name as `name` is already existing in self,
        return the already existing one. Otherwise, if `name` is a View
        itself, return `name`. If None of this is True, return None.

        Parameters
        ----------
        name : str or GraphMember

        Returns
        -------
        View or None
        """
        if name is None and group is None:
            return None
        return self.member(name=name, group=group, instance_class=View)

    @if_open
    def _modification(self, name=None, group=None):
        """
        If a Modification with the same name as `name` is already existing in
        self, return the already existing one. Otherwise, if `name` is a
        Modification itself, return `name`. If None of this is True, return
        None.

        Parameters
        ----------
        name : str or grap.GraphMember

        Returns
        -------
        Modification or None
        """
        if name is None and group is None:
            return None
        return self.member(name=name, group=group, instance_class=Modification)

    @if_open
    def print_status(self, notes=True, multiregion=False):
        """
        Prints a status of all Records, Views, and Modifications of this
        Experiment.
        """
        if notes and self.notes != "":
            print("Notes:")
            print(self.notes)
            print("-" * 119)
        print("name \t\t      group \t\tclass\t\t  caching/updated")
        print("-" * 119)
        if multiregion:
            for region in self.regions():
                print(str(region) + "\t"
                      + "c: " + str(region.caching))
        else:
            for record in self.records():
                print(str(record) + "\t"
                      + "c: " + str(record.caching))
            for view in self.views():
                print(str(view) + "\t"
                      + "c: " + str(view.caching))
        for modification in self.modifications():
            print(str(modification) + "\t"
                  + "u: " + str(modification.updated))

    @if_open
    def set_cached_region(self, region, group=None, update_cache=True):
        """
        Set the Region whose data should be cached and unset the caching of the
        previously cached Region (see attribute `self.cached_region`).

        Parameters
        ----------
        region : str or GraphMember or collections.Iterable of these
            Name of the Region, a GraphMember with the name, or the Region
            itself, whose caching should be switched on.
        group : str, optional
            The name of the group, Defaults to None or `region.group`.
        update_cache : bool
            Immidiately trigger an update of the cache of the new Region.
        """
        region = self.region(region, group=group)

        if not isinstance(region, collections.Iterable):
            region = [region]

        # switch on caching of this Region
        for rgn in region:
            rgn.caching = True
            if update_cache:
                rgn.update_cache()

        # switch off caching of previous Region
        if self.cached_region != region:
            for crgn in self.cached_region:
                if not isinstance(crgn, Record):
                    crgn.caching = False

        self.cached_region = region

    @if_open
    def has_group(self, group):
        """
        Check, if a group with the name `group` exist.

        A group can be either a View or a Modification with its View based.

        Parameters
        ----------
        group : str
            The name of the group.
        """
        return self._graphroot.group_root(group) is not None

    @if_open
    def save(self, pack=False, description=None, filename=None, directory=None,
             verbose=True):
        """
        Save the current status of the Experiment to file.

        Parameters
        ----------
        pack : bool, optional
            If pack is True, call `self.cleanup()`.
        description : str, optional
            Associate and stora a description along with all the changes from
            last till this saving point.
        filename : str, optional
            New filename to rename the experiment file to. Defaults to current
            filename. Only used if either `filename` or `directory` are set.
        directory : str, optional
            New directory to move the experiment file to. If tempfile, defaults
            to the current working directory. Otherwise, to the current
            directory of self.filename.
        """
        # Save the status of the experiment to file
        self._save(pack=pack, description=description)

        # Move the experiment file to a new location, if requested
        if filename is not None or directory is not None:
            self.move_to(filename=filename, directory=directory)

        # Warn user about the consequences, if temporary mode is switched on
        if self._tempdir is not None and verbose:
            print("ATTENTION: Experiment is in temporary mode, only. If you "
                  "close the experiment file, all work is lost! To switch off "
                  "the temporary mode and store the experiment file "
                  "permanently, either provide a `filename` and/or a "
                  "`directory`, or directly call the method `self.move_to` "
                  "prior to closing the experiment file.")

    @if_open
    def _save(self, pack=False, description=None):
        # Get current (or new) transaction
        txn = self._transaction_manager.get()

        # Set a description for the current transaction
        description = description or ""
        txn.note(description)
        # if sys.platform in ['linux', 'linux2', 'darwin']:
        #   doesn't work in notebook via internet
        #   user = os.getlogin()
        # else:
        # Set a user for the current transaction
        user = getpass.getuser()
        txn.setUser(user)

        # commit the transaction (i.e. save the experiment file)
        txn.commit()

        # Delete old undo data
        if pack:
            self.cleanup()

    @if_open
    def move_to(self, filename=None, directory=None):
        """
        Rename experiment file to a new name and/or move it to a now directory.

        Parameters
        ----------
        filename : str
            New filename to rename the experiment file to. Defaults to current
            filename.
        directory : str
            New directory to move the experiment file to. If tempfile, defaults
            to the current working directory. Otherwise, to the current
            directory of self.filename.

        Returns
        -------
        bool
        """
        # Get new and old database path
        old_filename, old_absdir, old_absfile = hp.file_and_dir(self.filename)
        if self._tempdir is None:
            # In non temp mode default directory to current one
            directory = directory or old_absdir
        filename, absdir, absfile = hp.file_and_dir(filename, directory)
        if filename is None:
            # If no filename was give, default to current filename
            filename, absdir, absfile = hp.file_and_dir(old_filename, absdir)

        if filename == old_filename and absdir == old_absdir:
            # No new filename and/or path, nothing to do, exit silently
            return False

        if not os.path.isdir(absdir):
            # directory where file should be stored does not exist
            print("A directory with the name '%s' does not exist." % absdir)
            return False

        if os.path.isfile(absfile):
            # another file with the same name already exists
            print("A file with the name '%s' already exists." % absfile)
            return False

        if os.path.isdir(absfile):
            # another directory with the same name already exists
            print("A directory with the name '%s' already exists." % absfile)
            return False

        # save current state of the experiment
        self._save()

        # Change ds._directory of all datasources to be relative to the new
        # absdir.
        changed = False
        for child in self._graphroot.children:
            if hasattr(child, 'datasource'):
                ds = child.datasource
                if ds._needs_file:
                    changed = True
                    ds._directory = os.path.relpath(ds.absdir, absdir)

        if changed:
            # save experiment with changed directories
            self._save(description="Change location of experiment to '%s'"
                       % absfile)

        # Close experiment file
        self.close(verbose=False, discard_temp=False)

        try:
            print("Move file from\n  '%s' to\n  '%s'"
                  % (old_absfile, absfile))
            # Move/rename experiment file to new location
            shutil.move(old_absfile, absfile)
        except:
            # File could not be moved, revert everything
            self.open(filename=old_absfile, verbose=False)
            if changed:
                self.undo(verbose=False)
            return False
        else:
            # Delete temporary ZODB files
            for ext in ['.lock', '.tmp', '.index', '.old']:
                try:
                    os.remove(''.join([old_absfile, ext]))
                except:
                    pass
            # delete tempdir if file was moved from tempdir
            self._rmtemp()

            # reopen the moved file
            self.open(filename=absfile, verbose=False)

        return True

    @if_open
    def abort(self):
        """
        Abort all changes of the Experiment back to the last saving point.
        """
        self._transaction_manager.abort()

    @if_open
    def cleanup(self):
        """
        Pack/clean the databasefile, the Experiment is stored in, to purge
        obsolete (undo) data.
        Changes of the database do not overwrite an existing object, they only
        create a new revision of that object. The benefit is, you may go back
        to an earlier revision, the drawback is, you have to pack the database
        from time to time to purge obsolete data.
        """
        self._db.pack()

    @if_open
    def undo(self, verbose=True):
        """
        Undo all changes of the Experiment back to the last saving point.
        """
        if not self._db.supportsUndo():
            if verbose:
                print("UNDO is not supported!")
            return

        # Get the last saved change
        undo_log = self._db.undoLog(0, -1)
        if len(undo_log) > 0:
            last_change = undo_log[0]
            if verbose:
                print("UNDOing: %s" % last_change)
            txn = self._transaction_manager.get()
            self._db.undo(last_change['id'], txn=txn)
            self._save(description="UNDO last change")
        else:
            if verbose:
                print('Nothing to UNDO.')

    @property
    def filename(self):
        """
        str
            The absolute path and filename of this experiment. It is either
            determined automatically upon initilization of the `Experiment`
            (temporary mode) or can alternatively be set by the parameters
            `filename` and `directory`.
        """
        return self._filename

    @property
    def is_open(self):
        """
        Check if experiment file is open.

        Returns
        -------
        bool
        """
        if '_connection' not in self.__dict__:
            return False
        return self._connection is not None

    @property
    @if_closed_return(True)
    def is_closed(self):
        """
        Check if experiment file is closed.

        Returns
        -------
        bool
        """
        return False

    @property
    @if_closed_return(dict)
    def _dbroot(self):
        return self._connection.root()

    @property
    @if_closed_return(version())
    def version(self):
        """
        str
            The version of the program (API and attributes) this experiment
            was created and saved.
        """
        return self._dbroot['version']

    @property
    @if_closed_return(dict)
    def _status(self):
        """
        picklable.Attributes
            self._status automatically stores any not yet existing attribute of
            the experiment. All stored attributes can be accessed via dotted
            notation and dict notation via self._status and dotted notation via
            experiment.

        Notes
        -----
        Make sure the attribute can be pickled and unpickled by ZODB/Python.
        """
        return self._dbroot['status']

    @property
    @if_closed_return()
    def notes(self):
        """
        Notes for this experiment.

        Returns
        -------
        str

        Examples
        --------
        To set the notes, assign the property `self.notes` as following:

        >>> self.notes = \"""
        >>> Put the multiline notes between the triple quotes.
        >>>
        >>> Empty lines are allowed.
        >>>
        >>> Empty lines at the beginning and the end of the text will be
        >>> stripped.
        >>> \"""
        """
        if 'notes' in self._status:
            return self._status.notes.strip()
        else:
            return ""

    @notes.setter
    @if_open
    def notes(self, notes):
        """
        Store notes for this experiment.

        Parameters
        ----------
        notes : str
            The notes to be stored
        """
        self._status.notes = notes

    @property
    @if_closed_return()
    def _graphroot(self):
        """
        GraphMember
            The root GraphMember, which holds the references (via a `Node`) to
            all Records of this experiment and subsequentially to all other
            Views and Modifications.
        """
        return self._dbroot['graphroot']

    def __getattr__(self, name):
        """
        Get the attributes that are stored in the `Attributes` status object
        of the Experiment.

        Parameters
        ----------
        name : str
            name is used as a key for self._status.
        """
        if self.is_open and name in self._status:
            return self._status[name]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        """
        Set an attribute that is stored in the `Attributes` status object of
        the Experiment.

        Parameters
        ----------
        name : str
            name is used as a key for self._status.
        value
            Object to be stored in self._status. Any picklable object will do.
        """
        # A __setattr__() hook will also override the Persistent __setattr__()
        # hook.  User code must treat it much like __getattribute__().  The
        # user-defined code must call _p_setattr() first to handle special
        # attributes; _p_setattr() takes the attribute name and value.  If it
        # returns True, Persistent handled the attribute. If not, the user code
        # can run. If the user code modifies the object’s state, it must be
        # assigned to _p_changed
        if (self.is_open  # Only set self._status.name variable if db is open
                          # 'self._connection' determines the open state.
                          # If you try to close the connection with
                          # "self._connection = None" in self.close(), after
                          # the connection was already closed, you have to make
                          # sure to not try to lookup '_connection' in
                          # self._status, which would fail, because the db is
                          # already closed, but self.is_open still returns a
                          # True.
                and name != '_connection'
                and name in self._status):
            self._status[name] = value
        else:
            super().__setattr__(name, value)

    def __str__(self):
        if self.filename is None:
            return self.__repr__()
        else:
            return ''.join([''.join(["'", self.filename, "':"]).ljust(20),
                            '<', self.__class__.__module__,
                            '.',
                            self.__class__.__name__, '>'])


def create_experiment(return_last_created=True, **kwargs):
    """
    Create a new Experiment or get the last created one.

    Parameters
    ----------
    return_last_created : bool
        If an experiment was already created in the python kernel instance,
        return that experiment instead of creating a new one.
    **kwargs
        Parameters for the initialization of an instance of Experiment().

    Returns
    -------
    Experiment
        The new or the last created Experiment.
    """
    if return_last_created:
        return Experiment.created_last or Experiment(**kwargs)
    else:
        return Experiment(**kwargs)


def open_experiment(filename=None, directory=None, return_last_created=True,
                    **kwargs):
    """
    Create a new Experiment or get the last created one.
    Optionally, open an experiment file `filename'.

    Parameters
    ----------
    filename : str, optional
        The path to the file of the Experiment.
    directory : str, optional
        If `filename` alone is not sufficient to describe the location of the
        experiment file, you can provide a `directory`, where the path of
        `filename` is located in.
    return_last_created : bool
        If an experiment was already created in the python kernel instance,
        return that experiment instead of creating a new one.
    **kwargs
        Parameters for the initialization of an instance of Experiment().

    Returns
    -------
    Experiment
        The Experiment with the opened file.
    """
    experiment = create_experiment(return_last_created=return_last_created,
                                   **kwargs)
    experiment.open(filename=filename, directory=directory)
    return experiment


def close_experiment():
    """
    Close the last created Experiment.
    """
    experiment = Experiment.created_last
    if experiment is not None:
        return experiment.close()
    return True


def save_experiment(**kwargs):
    """
    Save the last created Experiment.
    """
    experiment = Experiment.created_last
    if experiment is not None:
        return experiment.save(**kwargs)
    return True


def create_calibration(calibration_type=None, verbose=False, **kwargs):
    """
    Create a Calibration object.

    Parameters
    ----------
    calibration_type : str
        Select the type of the Calibration. Possible types can be found in the
        default `cfgfile` `pyoti/etc/calibration.cfg`
    verbose : bool
        Be verbose about the determination of the calibration_type upon reading
        the paramaters from the `cfgfile`.
    **kwargs
        Parameters will be passed to the function
        `calibration.create_calibration()`

    Returns
    -------
    Calibration
        A Calibration object.
    """
    calibration = cb.create_calibration(source_type=calibration_type,
                                        verbose=verbose, **kwargs)
    print('Created calibration with: ')
    print('  object: ' + str(calibration.calibsource))
    print('  name: ' + calibration.calibsource.name)
    print('  beta: ' + str(calibration.beta()))
    print('  kappa: ' + str(calibration.kappa()))
    print('If this seems suspicious to you, recheck the parameters you '
          'provided!')
    return calibration



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 13:36:39 2016

@author: Tobias Jachowski
"""
import matplotlib.pyplot as plt
import numpy as np

from . import signal as sn
from .. import helpers as hp
from .evaluate import Evaluator
from .signalfeature import CycleSectioner


class Tether(Evaluator):
    """
    Calculates tether related parameters, like sections of stressing or
    releasing the bead, bead-center surface distance, extension of the tether,
    or force acting on the tether.

    To distinguish between the excited axes ('x', 'y'), directions
    ('left', 'right'), and different cycles ('stress', 'release'), a
    CycleSectioner is used.
    """
    def __init__(self, region=None, calibration=None, resolution=1000,
                 filter_time=-1.0, resolution_sf=None, filter_time_sf=None,
                 **kwargs):
        """
        This __init__() constructor extends the superclass (Evaluator)
        constructor by setting the traces_sf to default to 'positionXY'.

        For extensive documentation of the Parameters and additional
        attributes, see `pyoti.evaluate.evaluate.Evaluator`.

        Parameters
        ----------
        region : pyoti.region.region.Region
        calibration : pyoti.calibration.calibration.Calibration
        resolution : float
        filter_time : float
        resolution_sf : float
        filter_time_sf : float
        **kwargs
            Extra arguments, e.g. traces_sf, in case the positionXY signals are
            called differently.

        Attributes
        ----------
        displacementXYZ : 2D numpy.ndarray
        forceXYZ : 2D numpy.ndarray
        force
        distanceXYZ
        distance
        extension
        angle
        rightstress
        leftstress
        leftrelease
        rightrelease
        stress
        release
        rightstresses
        leftstresses
        leftreleases
        rightreleases
        stresses
        releases
        """
        sf_class = CycleSectioner
        traces_sf = kwargs.pop('traces_sf', 'positionXY')

        # resolution for (almost) all properties, e.g. displacement, force, ...
        # resolution 1.000 Hz, samplingrate 40.000 Hz -> 40 points
        super().__init__(region=region, calibration=calibration,
                         resolution=resolution, filter_time=filter_time,
                         sf_class=sf_class, traces_sf=traces_sf,
                         resolution_sf=resolution_sf,
                         filter_time_sf=filter_time_sf)

        self.fe_figure = None

    def _sections(self, axis=None, direction=None, cycle=None,
                  concat_direction=False, concat_cycle=False, info=False,
                  **kwargs):
        """
        Calculate start/stop indices (segments) of the requested axis,
        direction, and cycle (sections).

        Parameters
        ----------
        axis : str or list of str, optional
            'x', 'y', or ['x', 'y'] (default)
        direction : str or list of str, optional
            'left', 'right', or ['left', 'right'] (default)
        cycle : str or list of str, optional
            'stress', 'release', or ['stress', 'release'] (default)
        concat_direction : bool, optional
            Concatenate individual left/right sections into one section. Only
            evaluated, if `concat_cycle` is True.
        concat_cycle : bool, optional
            Concatenate individual stress/release sections into one new section
            with the start beeing the start of stress and the stop beeing the
            stop of release.
        info : bool, optional
            Additionally to the sections return a 2D numpy.ndarray, containing
            the axis, direction, and the cycle for every individual section.
            Axis can be either 'x' or 'y'. Direction can be 'left', 'right' or
            'leftright'. Cycle can be 'stress', 'release', or 'stressrelease'.

        Returns
        -------
        2D np.ndarray or tuple of two 2D np.ndarray
            If info is False start/stop values in rows.
            If info is True start/stop values in rows and an 2D np.ndarray with
            infos (see parameter `info`).
        """
        if axis is None:
            axis = ['x', 'y']
        if direction is None:
            direction = ['left', 'right']
        else:
            # A direction was chosen, prevent concat
            concat_direction = False
        if isinstance(direction, str):
            direction = [direction]
        if cycle is None:
            cycle = ['stress', 'release']
        else:
            # A cycle was chosen, prevent concat
            concat_cycle = False
        if isinstance(cycle, str):
            cycle = [cycle]

        sections = np.empty((0, 2), dtype=int)
        infos = np.empty((0, 3), dtype=str)

        # Get all requested segments
        # for ax, axi in zip(axis, range(len(axis))):
        for ax in axis:
            # separate release/stress and separate left/right -> x/y,
            # right/left, stress/release
            if not concat_cycle:
                for sec, dc in zip([d + c for d in direction for c in cycle],
                                   [[d, c] for d in direction for c in cycle]):
                    for segment in self._sf.sections[ax]:
                        sections = np.r_[sections, segment[sec]]
                        if len(segment[sec]) > 0:
                            seg_info = np.array([[ax, dc[0], dc[1]]]
                                                * len(segment[sec]))
                            infos = np.r_[infos, seg_info]
            # concat release/stress and separate left/right -> x/y, right/left
            if concat_cycle and not concat_direction:
                # for d, di in zip(direction, range(len(direction))):
                for d in direction:
                    for segment in self._sf.sections[ax]:
                        sections = np.r_[sections, segment[d]]
                        if len(segment[d]) > 0:
                            seg_info = np.array([[ax, d, 'stressrelease']]
                                                * len(segment[d]))
                            infos = np.r_[infos, seg_info]
            # concat release/stress and concat left/right -> x/y
            if concat_cycle and concat_direction:
                sections = np.r_[sections, self._sf.excited[ax]]
                if len(self._sf.excited[ax]) > 0:
                    seg_info = np.array([[ax, 'leftright', 'stressrelease']]
                                        * len(self._sf.excited[ax]))
                    infos = np.r_[infos, seg_info]

        if info:
            return sections, infos
        else:
            return sections

    def _extrema(self, axis=None, extremum=None):
        if axis is None:
            axis = ['x', 'y']
        if extremum is None:
            extremum = ['minima', 'maxima']
        elif isinstance(extremum, str):
            extremum = [extremum]

        extrema = np.empty(0, dtype=int)

        # Get all extrema and sort
        for ax in axis:
            for section in self._sf.sections[ax]:
                for ext_type in extremum:
                    idx = section[ext_type]
                    idx = self.undecimate_and_limit(idx)
                    extrema = np.r_[extrema, idx]
        extrema.sort()
        return extrema

    def stress_release_pairs(self, axis=None, direction=None, i=None,
                             slices=True, decimate=None, info=False, **kwargs):
        """
        Calculate start/stop indices (as segments or slices) of stress/release
        cycle pairs of the requested axis and direction.

        Parameters
        ----------
        axis : str or list of str, optional
            'x', 'y', or ['x', 'y'] (default)
        direction : str or list of str, optional
            'left', 'right', or ['left', 'right'] (default)
        i : int, optional
            The index of the stress release pair to be selected and returned
            from all stress release pairs calculated. Depending on `axis` and
            `direction` the total availabe number of stress release pairs may
            vary.
        slices : bool, optional
            Set to False to return segments instead of slices. Default is True.
        decimate : int, optional
            Used to set the step attribute of the returned slices. Only
            evaluated, if `slices` is True.
        info : bool, optional
        **kwargs : dict, optional
            Only used for compatibility purposes, to be able to call method
            with parameters not defined in method definition.

        Returns
        -------
        stresses : 1D np.ndarray of slices or segments
        releases : 1D np.ndarray of slices or segments
        stress_infos : 1D np.ndarray of 1D np.ndarrays of type str
            The str arrays have the form of [axis, direction, cycle]. Only
            returned, if parameter `info` is True.
        release_infos : 1D np.ndarray of 1D np.ndarrays of type str
            The str arrays have the form of [axis, direction, cycle]. Only
            returned, if parameter `info` is True.
        """
        if axis is None:
            axis = ['x', 'y']
        if direction is None:
            direction = ['left', 'right']
        if isinstance(direction, str):
            direction = [direction]
        stress_segments, _stress_infos = self.sections(axis=axis,
                                                       direction=direction,
                                                       cycle='stress',
                                                       slices=False,
                                                       info=True)
        release_segments, _release_infos = self.sections(axis=axis,
                                                         direction=direction,
                                                         cycle='release',
                                                         slices=False,
                                                         info=True)
        extrema = self._extrema(axis=axis)
        stresses = np.empty((0, 2), dtype=int)
        releases = np.empty((0, 2), dtype=int)
        stress_infos = np.empty((0, 3), dtype=str)
        release_infos = np.empty((0, 3), dtype=str)

        # Group all stress/release cycle pairs according to the extrema
        # A stress release pair corresponds to one extremum, only if the stop
        # of the stress and the start of the release segment equal the
        # extremum. However, either one of the stress or release segment can be
        # missing.
        for ext in extrema:
            # Find stress whose stop is equal to extremum
            stress_idx = stress_segments[:, 1] == ext
            stress = stress_segments[stress_idx]
            stress_info = _stress_infos[stress_idx]
            # Find release whose start is equal to extremum
            release_idx = release_segments[:, 0] == ext
            release = release_segments[release_idx]
            release_info = _release_infos[release_idx]
            # Create stress slice, if stress is empty, but release is valid
            if stress.size == 0 and release.size > 0:
                stress = np.array([[ext, ext]])
            # Create release slice, if release is empty, but stress is valid
            if release.size == 0 and stress.size > 0:
                release = np.array([[ext, ext]])
            if stress.size > 0 and release.size > 0:
                # p_p = np.r_[stress, release]  # concatenate stress/release
                # cycle pair
                # concatenate with other cycles
                stresses = np.r_[stresses, stress]
                releases = np.r_[releases, release]
                stress_infos = np.r_[stress_infos, stress_info]
                release_infos = np.r_[release_infos, release_info]

        # Convert segments into slices
        if slices:
            stresses = sn.idx_segments_to_slices(stresses, decimate=decimate)
            releases = sn.idx_segments_to_slices(releases, decimate=decimate)

        # Get the maximum number of stress release pairs
        stop = len(stresses)
        # Prevent index overflow and allow negative indices
        if i is not None:
            if i < 0:
                i = stop + i
            i = max(0, i)
            i = min(i, stop - 1)
            s = slice(i, i + 1)
        else:
            s = slice(0, stop)

        if info:
            return stresses[s], releases[s], stress_infos[s], release_infos[s]
        else:
            return stresses[s], releases[s]

    def baseline_idx(self, axis=None, strict=False, extrapolate=False):
        """
        Calculate the indices of the baseline, according to the excited axes,
        i.e. the indices, where the excited axis is zero. This method uses
        the detected sections of `self.region`, instead of searching for
        the value 0.0 in the excited axes, as compared to the function
        `pyoti.evaluate.signal.basline_idx()`.

        Parameters
        ----------
        axis : str or list of str, optional
            'x', 'y', or ['x', 'y'] (default)
        strict : bool
            Take only baseline indices of releases's stops or stresses's
            starts, if a stress directly follows a release section.
        extrapolate : bool
            Take baseline indices, even if they are indexed by only one
            stress or release segment, i.e. indices which mark the start or
            the end of a wave.
        """
        # Get stress/release pairs
        if strict:
            # Get only stress/release pairs, i.e. only segment pairs, whose
            # release's stop and stress's start index equal an extremum.
            stresses, releases = self.stress_release_pairs(axis=axis,
                                                           slices=False)
        else:
            # Get all stress/release segments, even those without a
            # corresponding extremum.
            stresses = self.sections(axis=axis, cycle='stress', slices=False)
            releases = self.sections(axis=axis, cycle='release', slices=False)

        # A baseline point lies exactly beetween a release and a following
        # stress section. Therefore, the baseline point idx equals a stress's
        # start and a release's stop.
        stress_base_idx = stresses[:, 0]
        release_base_idx = releases[:, 1]

        # Remove stresses, whose start, and releases, whose stop is equal to
        # an extremum, which means, the excitation as above or below the
        # baseline.
        extrema = self._extrema(axis=axis)
        stress_base_idx = np.setdiff1d(stress_base_idx, extrema,
                                       assume_unique=True)
        release_base_idx = np.setdiff1d(release_base_idx, extrema,
                                        assume_unique=True)

        base_idx = np.r_[stress_base_idx, release_base_idx]
        if strict or not extrapolate:
            # Take only baseline point indices, that lie exactly between a
            # release and a following stress section, i.e. have two equal
            # entries, one from a release's start and one from a stress's stop.
            # I.e. sort out all indices that did not come from a stress/release
            # pair but instead only a single stress or release segment.
            base_idx.sort()
            base_idx = base_idx[np.r_[base_idx[:-1] == base_idx[1:], False]]
        return np.unique(base_idx)

    def _rfigure(self, legend=True, fig=None, ax=None):
        """
        Plot the overview of detected stress release cycles.
        """
        if fig is None and ax is None:
            fig, ax = plt.subplots()
            suptitle = True
        elif fig is None:
            fig = ax.get_figure()
            suptitle = False
        elif ax is None:
            ax = fig.gca()
            suptitle = False

        ax.grid(True)

        line_rstr = None
        line_rrls = None
        line_lstr = None
        line_lrls = None
        line_minima = None
        line_maxima = None
        t = self.timevector
        for axis, trace in zip('xy', ['positionX', 'positionY']):
            s = self.get_data(traces=trace) * 1e6  # m -> µm
            rstr, rrls = self.stress_release_pairs(axis=axis,
                                                   direction='right')
            lstr, lrls = self.stress_release_pairs(axis=axis, direction='left')

            ax.plot(t, s, lw=0.1, ms=2, color='k', alpha=1.0)

            # line_rstr = None
            # line_rrls = None
            # line_lstr = None
            # line_lrls = None
            for rstr, rrls in zip(rstr, rrls):
                line_rstr, = ax.plot(t[rstr], s[rstr], lw=0.4, ms=2, color='m')
                line_rrls, = ax.plot(t[rrls], s[rrls], lw=0.4, ms=2, color='c')
            for lstr, lrls in zip(lstr, lrls):
                line_lstr, = ax.plot(t[lstr], s[lstr], lw=0.4, ms=2, color='g')
                line_lrls, = ax.plot(t[lrls], s[lrls], lw=0.4, ms=2, color='y')

            # line_minima = None
            # line_maxima = None
            for segment in self._sf.sections[axis]:
                minima = self.undecimate_and_limit(segment['minima'])
                maxima = self.undecimate_and_limit(segment['maxima'])
                line_minima, = ax.plot(t[minima], s[minima], '.', ms=5,
                                       color='b')
                line_maxima, = ax.plot(t[maxima], s[maxima], '.', ms=5,
                                       color='r')

        line_excited_x = None
        for x_c in (self.undecimate_and_limit(self._sf.excited['x'])
                    / self.resolution):
            line_excited_x = ax.hlines(0.0, x_c[0], x_c[1], alpha=1,
                                       colors='b', linestyle='solid', lw=1)
            # ax.plot(x_c[0], 0.5, '.k', alpha=1, ms=3)
            # ax.plot(x_c[1], 0.5, '.k', alpha=1, ms=3)
            ax.vlines(x_c[0], -0.01, 0.01, alpha=1, colors='b',
                      linestyle='solid', lw=1)
            ax.vlines(x_c[1], -0.01, 0.01, alpha=1, colors='b',
                      linestyle='solid', lw=1)

        line_excited_y = None
        for y_c in (self.undecimate_and_limit(self._sf.excited['y'])
                    / self.resolution):
            line_excited_y = ax.hlines(0.0, y_c[0], y_c[1], alpha=1,
                                       colors='r', linestyle='solid', lw=1)
            # ax.plot(y_c[0], -0.5, '.k', alpha=1, ms=3)
            # ax.plot(y_c[1], -0.5, '.k', alpha=1, ms=3)
            ax.vlines(y_c[0], -0.01, 0.01, alpha=1, colors='r',
                      linestyle='solid', lw=1)
            ax.vlines(y_c[1], -0.01, 0.01, alpha=1, colors='r',
                      linestyle='solid', lw=1)

        ax.set_xlim((t[0], t[-1]))

        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Signal positionX and Y (µm)")
        if suptitle:
            fig.suptitle("Automatically detected excited axis, minima, "
                         "maxima, and sections.")

        if legend:
            if line_minima is not None:
                line_minima.set_label('minima')
            if line_maxima is not None:
                line_maxima.set_label('maxima')
            if line_rstr is not None:
                line_rstr.set_label('rightstress')
            if line_rrls is not None:
                line_rrls.set_label('rightrelease')
            if line_lstr is not None:
                line_lstr.set_label('leftstress')
            if line_lrls is not None:
                line_lrls.set_label('leftrelease')
            if line_excited_x is not None:
                line_excited_x.set_label('excited x')
            if line_excited_y is not None:
                line_excited_y.set_label('excited y')

            ax.legend(loc='upper right')

        return fig

    def force_extension_pair(self, i, axis=None, direction=None, decimate=None,
                             time=False, twoD=False, posmin=10e-9):
        """
        Calculate the force extension pair with index `i`.

        Parameters
        ----------
        i : int
            Index of the force_extension pair to be returned.
        axis : str
            See method `self.stress_release_pairs()`.
        direction : str
            See method `self.stress_release_pairs()`.
        decimate : int
            See method `self.stress_release_pairs()`.
        posmin : float
            The `posmin` is used to decide wether the magnitude of the force
            has to be corrected with the sign depending on the position. The
            `posmin` sets the value the position signal has to be deflected at
            least to be counted as active pulling on the bead. The value should
            at least be >= 12 times the standard deviation of the unexcited
            position signal.
            Smaller values could (depending on the number of datapoints)
            possibly lead to falsly detected excitation of the signal.

        Returns
        -------
        1D numpy.ndarray of type float
            Extension values of stress cycles in m.
        1D numpy.ndarray of type float
            Force values of stress cycles in N.
        1D numpy.ndarray of type str
            (str, str, str), containing the axis, direction, and the cycle.
            Axis can be either 'x' or 'y'. Direction can be 'left', or 'right'.
            Cycle is 'stress'.
        1D numpy.ndarray of type float
            Extension values of release cycles in m.
        1D numpy.ndarray of type float
            Force values of release cycles in N.
        1D numpy.ndarray of type str
            (str, str, str), containing the axis, direction, and the cycle.
            Axis can be either 'x' or 'y'. Direction can be 'left', or 'right'.
            Cycle is 'release'.
        """
        fe_pair = self.force_extension_pairs(axis=axis, direction=direction,
                                             i=i, decimate=decimate, time=time,
                                             twoD=twoD, posmin=posmin)
        return next(fe_pair)

    def force_extension_pairs(self, axis=None, direction=None, i=None,
                              decimate=None, time=False, twoD=False,
                              posmin=10e-9):
        """
        Return a generator for force extension values of stress release pairs.

        Parameters
        ----------
        str_rls_pairs : tuple of 4 numpy.ndarrays, optional
        axis : str, optional
            See method `self.stress_release_pairs()`.
        direction : str, optional
            See method `self.stress_release_pairs()`.
        i : int
            Index of the force extension pair to be yielded.
        decimate : int, optional
            See method `seld.stress_release_pairs()`.
        posmin : float
            The `posmin` is used to decide wether the magnitude of the force
            has to be corrected with the sign depending on the position. The
            `posmin` sets the value the position signal has to be deflected at
            least to be counted as active pulling on the bead. The value should
            at least be >= 12 times the standard deviation of the unexcited
            position signal.
            Smaller values could (depending on the number of datapoints)
            possibly lead to falsly detected excitation of the signal.

        Yields
        ------
        1D numpy.ndarray of type float
            Extension values of stress cycles in m.
        1D numpy.ndarray of type float
            Force values of stress cycles in N.
        1D numpy.ndarray of type str
            (str, str, str), containing the axis, direction, and the cycle.
            Axis can be either 'x' or 'y'. Direction can be 'left', or 'right'.
            Cycle is 'stress'.
        1D numpy.ndarray of type float
            Extension values of release cycles in m.
        1D numpy.ndarray of type float
            Force values of release cycles in N.
        1D numpy.ndarray of type str
            (str, str, str), containing the axis, direction, and the cycle.
            Axis can be either 'x' or 'y'. Direction can be 'left', or 'right'.
            Cycle is 'release'.
        """
        str_rls_pairs = self.stress_release_pairs(axis=axis,
                                                  direction=direction,
                                                  i=i,
                                                  decimate=decimate,
                                                  slices=True,
                                                  info=True)
        strs, rlss, stri, rlsi = str_rls_pairs

        # Get the start from the first stress and the stop from the last
        # release cycle
        start = strs[0].start
        stop = rlss[-1].stop
        samples = slice(start, stop)

        # Get extension, force, and stress/release pairs
        e_f = self.force_extension(samples=samples, twoD=twoD, posmin=posmin)  # m,N
        e = e_f[:, 0]
        f = e_f[:, 1]

        # Get the time
        if time:
            t = self.timevector[samples]

        # Yield all stress/release extension/force data pairs
        for st, rl, sti, rli in zip(strs, rlss, stri, rlsi):
            _st = slice(st.start - start, st.stop - start, st.step)
            _rl = slice(rl.start - start, rl.stop - start, rl.step)
            est = e[_st]
            fst = f[_st]
            erl = e[_rl]
            frl = f[_rl]
            if time:
                tst = t[_st]
                trl = t[_rl]
                yield est, fst, sti, erl, frl, rli, tst, trl
            else:
                yield est, fst, sti, erl, frl, rli

    def samples(self, i, cycle=None, axis=None, direction=None, decimate=None):
        """
        Get index of samples to be used for functions displacement, force, etc.

        Parameters
        ----------
        cycle : str or list of str
            cycle can be either 'stress' or 'release'
        """
        if cycle is None:
            cycle = ['stress', 'release']
        if 'stress' in cycle and not 'release' in cycle:  # stress only
            start = 0
            stop = 0
        elif 'release' in cycle and not 'stress' in cycle:  # release only
            start = 1
            stop = 1
        else:  # stress and release
            start = 0
            stop = 1

        str_rls_pair = self.stress_release_pairs(axis=axis,
                                                 direction=direction,
                                                 i=i,
                                                 decimate=decimate,
                                                 slices=False,
                                                 info=False)
        idx_start = str_rls_pair[start][0][0]
        idx_stop = str_rls_pair[stop][0][1]
        idx = slice(idx_start, idx_stop, decimate)
        return idx

    def displacementXYZ(self, samples=None):
        """
        Displacement in m with height dependent calibration factors for X, Y
        and Z.
        """
        # Get extension (in a fast way)
        data = self.get_data(traces=['psdXYZ', 'positionZ'], samples=samples)
        psdXYZ = data[:, 0:3]
        positionZ = data[:, [3]]
        calibration = self.calibration

        dispXYZ = displacementXYZ(calibration, psdXYZ, positionZ)
        return dispXYZ

    def forceXYZ(self, samples=None):
        """
        Force in N, that is acting on the tether
        """
        data = self.get_data(traces=['psdXYZ', 'positionZ'], samples=samples)
        psdXYZ = data[:, 0:3]
        positionZ = data[:, [3]]
        calibration = self.calibration

        fXYZ = forceXYZ(calibration, psdXYZ, positionZ)
        return fXYZ

    def force(self, samples=None, twoD=False, posmin=10e-9):
        """
        Magnitude of the force in N acting on the tethered molecule (1D
        numpy.ndarray).

        Parameters
        ----------
        posmin : float
            The `posmin` is used to decide wether the magnitude of the force
            has to be corrected with the sign depending on the position. The
            `posmin` sets the value the position signal has to be deflected at
            least to be counted as active pulling on the bead. The value should
            at least be >= 12 times the standard deviation of the unexcited
            position signal.
            Smaller values could (depending on the number of datapoints)
            possibly lead to falsly detected excitation of the signal.
        """
        # Get force (in a fast way)
        data = self.get_data(traces=['psdXYZ', 'positionXYZ'], samples=samples)
        psdXYZ = data[:, 0:3]
        positionXY = data[:, 3:5]
        positionZ = data[:, [5]]
        calibration = self.calibration

        # 2D or 3D calculation of the distance in Z
        if twoD:
            psdXYZ[:, Z] = 0.0

        fXYZ = forceXYZ(calibration, psdXYZ, positionZ)
        f = force(fXYZ, positionXY, posmin=posmin)
        return f

    def distanceXYZ(self, samples=None):
        """
        Distance of the attachment point to the bead center as a 3D vector.
        """
        data = self.get_data(traces=['psdXYZ', 'positionXYZ'], samples=samples)
        psdXYZ = data[:, 0:3]
        positionXYZ = data[:, 3:6]
        calibration = self.calibration

        distXYZ = distanceXYZ(calibration, psdXYZ, positionXYZ)
        return distXYZ

    def distance(self, samples=None, twoD=False, posmin=10e-9):
        """
        Distance of the attachment point to the bead center.

        Parameters
        ----------
        posmin : float
            The `posmin` is used to decide wether the magnitude of the force
            has to be corrected with the sign depending on the position. The
            `posmin` sets the value the position signal has to be deflected at
            least to be counted as active pulling on the bead. The value should
            at least be >= 12 times the standard deviation of the unexcited
            position signal.
            Smaller values could (depending on the number of datapoints)
            possibly lead to falsly detected excitation of the signal.
        """
        data = self.get_data(traces=['psdXYZ', 'positionXYZ'], samples=samples)
        psdXYZ = data[:, 0:3]
        positionXYZ = data[:, 3:6]
        positionXY = data[:, 3:5]
        calibration = self.calibration

        # 2D or 3D calculation of the distance in Z
        if twoD:
            psdXYZ[:, Z] = 0.0

        distXYZ = distanceXYZ(calibration, psdXYZ, positionXYZ)
        dist = distance(distXYZ, positionXY, posmin=posmin)
        return dist

    def extension(self, samples=None, twoD=False, posmin=10e-9):
        """
        Extension of the tethered molecule in m.

        Parameters
        ----------
        posmin : float
            The `posmin` is used to decide wether the magnitude of the force
            has to be corrected with the sign depending on the position. The
            `posmin` sets the value the position signal has to be deflected at
            least to be counted as active pulling on the bead. The value should
            at least be >= 12 times the standard deviation of the unexcited
            position signal.
            Smaller values could (depending on the number of datapoints)
            possibly lead to falsly detected excitation of the signal.
        """
        calibration = self.calibration

        dist = self.distance(samples=samples, twoD=twoD, posmin=posmin)
        e = extension(dist, calibration.radius)
        return e

    def force_extension(self, samples=None, twoD=False, posmin=10e-9):
        """
        Extension (m, first column) of and force (N, second column) acting
        on the tethered molecule (2D numpy.ndarray).

        Parameters
        ----------
        posmin : float
            The `posmin` is used to decide wether the magnitude of the force
            has to be corrected with the sign depending on the position. The
            `posmin` sets the value the position signal has to be deflected at
            least to be counted as active pulling on the bead. The value should
            at least be >= 12 times the standard deviation of the unexcited
            position signal.
            Smaller values could (depending on the number of datapoints)
            possibly lead to falsly detected excitation of the signal.
        """
        # Get extension and force (in a fast way)
        data = self.get_data(traces=['psdXYZ', 'positionXYZ'], samples=samples)
        psdXYZ = data[:, 0:3]
        positionXYZ = data[:, 3:6]
        positionXY = positionXYZ[:, 0:2]
        positionZ = data[:, [5]]
        calibration = self.calibration

        # 2D or 3D calculation of the distance in Z
        if twoD:
            psdXYZ[:, Z] = 0.0
        distXYZ = distanceXYZ(calibration, psdXYZ, positionXYZ)
        dist = distance(distXYZ, positionXY, posmin=posmin)
        e = extension(dist, calibration.radius)

        fXYZ = forceXYZ(calibration, psdXYZ, positionZ)
        f = force(fXYZ, positionXY, posmin=posmin)
        return np.c_[e, f]

    def info(self, i=0):
        print_info(self, i=i)

    @property
    def angle(self):
        """
        Returns a dictionary of angles, for three corners (A), (B), and (C),
        calculated by (F)orce and (D)istance:
            B
            |\
          a | \ c
            |  \
            |___\
           C  b  A
        """
        return angle(self.force, self.forceXYZ, self.excited_axis,
                     self.distance, self.distanceXYZ)

    @property
    def rightstress(self):
        return self.sections(direction='right', cycle='stress',
                             range_concat=True)

    @property
    def leftstress(self):
        return self.sections(direction='left', cycle='stress',
                             range_concat=True)

    @property
    def leftrelease(self):
        return self.sections(direction='left', cycle='release',
                             range_concat=True)

    @property
    def rightrelease(self):
        return self.sections(direction='right', cycle='release',
                             range_concat=True)

    @property
    def stress(self):
        return self.sections(cycle='stress', range_concat=True)

    @property
    def release(self):
        return self.sections(cycle='release', range_concat=True)

    @property
    def rightstresses(self):
        return self.sections(direction='right', cycle='stress')

    @property
    def leftstresses(self):
        return self.sections(direction='left', cycle='stress')

    @property
    def leftreleases(self):
        return self.sections(direction='left', cycle='release')

    @property
    def rightreleases(self):
        return self.sections(direction='right', cycle='release')

    @property
    def stresses(self):
        return self.sections(cycle='stress')

    @property
    def releases(self):
        return self.sections(cycle='release')


# Define constants for convenient handling
X = 0
Y = 1
Z = 2
XY = hp.slicify([X, Y])
XZ = hp.slicify([X, Z])
YZ = hp.slicify([Y, Z])
XYZ = hp.slicify([X, Y, Z])


def print_info(tether, i=0):
    stress_release_pair = tether.stress_release_pairs(i=i, info=True)
    stress, release, stress_info, release_info = stress_release_pair
    resolution = tether.resolution
    start_stress_t = stress[0].start / resolution
    stop_stress_t = stress[0].stop / resolution
    start_release_t = release[0].start / resolution
    stop_release_t = release[0].stop / resolution
    print("Stress release pair: #{:03d}".format(i))
    print("Focal shift: {:.3f}".format(tether.calibration.focalshift))
    print("Axis: {}".format(stress_info[0][0]))
    print("  stress")
    print("    t:  {:.3f} s".format(stop_stress_t - start_stress_t))
    print("        {:.2f} s - {:.2f} s".format(start_stress_t, stop_stress_t))
    print("    datapoints: {}".format(stress[0].stop - stress[0].start))
    print("    z0: {:.2f} nm".format(
        - np.median(tether.get_data('positionZ', stress[0])) * 1e9))
    print("    h0: {:.2f} nm".format(
        - np.median(tether.get_data('positionZ', stress[0])) * 1e9
                                    * tether.calibration.focalshift))
    print("  release")
    print("    t:  {:.3f} s".format(stop_release_t - start_release_t))
    print("        {:.2f} s - {:.2f} s".format(start_release_t,
                                               stop_release_t))
    print("    datapoints: {}".format(release[0].stop - release[0].start))
    print("    z0: {:.2f} nm".format(
        - np.median(tether.get_data('positionZ', release[0])) * 1e9))
    print("    h0: {:.2f} nm".format(
        - np.median(tether.get_data('positionZ', release[0])) * 1e9
                                    * tether.calibration.focalshift))


def displacementXYZ(calibration, psdXYZ, positionZ=0.0):
    """
    Displacement in m with height dependent calibration factors for X, Y
    and Z.
    """
    dispXYZ = calibration.displacement(psdXYZ, positionZ=positionZ)
    return dispXYZ


def forceXYZ(calibration, psdXYZ, positionZ):
    """
    Force acting on the bead
    """
    displacementXYZ = calibration.displacement(psdXYZ, positionZ=positionZ)
    fXYZ = calibration.force(displacementXYZ, positionZ=positionZ)

    # A displacement of the bead in the positive (negative) direction results
    # in a force acting on the bead in the opposite negative (positive)
    # direction.
    fXYZ *= - 1.0

    return fXYZ


def force(forceXYZ, positionXY, posmin=10e-9):
    """
    Parameters
    ----------
    forceXYZ : 2D numpy.ndarray of type float
        forceXYZ.shape[1] can consist of either 3 (XYZ) or 2 (XY) axes
    posmin : float
        The `posmin` is used to decide wether the magnitude of the force has to
        be corrected with the sign depending on the position. The `posmin` sets
        the value the position signal has to be deflected at least to be
        counted as active pulling on the bead. The value should at least be >=
        12 times the standard deviation of the unexcited position signal.
        Smaller values could (depending on the number of datapoints) possibly
        lead to falsly detected excitation of the signal.
    """
    if posmin is None:
        posmin = 0
    # The sign of the magnitude, i.e. the direction of the force is important
    # for the noise around +/- 0 N.
    # The sign of the magnitude of the forceXY depends on the positionXY. If we
    # pull the bead to one side and the bead is displaced to the same side, we
    # get a positive magnitude of the force. If the bead is displaced to the
    # opposite side of the one we are pulling to, we get a negative magnitude
    # of the force. Keep in mind that a positive displacement results in an
    # opposite directed negative force acting on the bead and vice versa. The
    # sign of the magnitude of the forceZ is independent of the positionZ.
    signF = np.sign(forceXYZ)
    # The position determines the direction of the force only of the axes where
    # we actively pull on the bead
    posmax = np.max(np.abs(positionXY), axis=0)
    idx = posmax >= posmin
    signF[:, XY][:, idx] *= - np.sign(positionXY[:, idx])

    # Square the forces and account for the signs
    force_sq = forceXYZ**2
    force_sq_sum = np.sum(force_sq, axis=1)
    # Calculate a "weighted" sign. Greater forces have greater influence on the
    # final sign
    sign_force_sum = np.sign(np.sum(force_sq * signF, axis=1))
    force = np.sqrt(force_sq_sum) * sign_force_sum
    return force


def distanceXYZ(calibration, psdXYZ, positionXYZ, radius=None, focalshift=None,
                clip_Z=True):
    """
    Distance of the attachment point to the bead center as a 3D vector.
    positionXYZ, displacementXYZ and radius need to have the same unit.
    If radius is 0.0, the positionZ is not corrected by the radius. You should
    set radius to 0.0, if the positionZ is defined in such a way, that
    positionZ is 0.0, where the bead center would be on the glass surface.
    If focalshift is 1.0, the positionXYZ is not corrected by the focalshift.

    Parameters
    ----------
    positionXYZ : 2D np.array of type float
    displacementXYZ : 2D np.array of type float
    radius : float
    focalshift : float
    clip_Z : bool
        The distance of the attachment point to the center of the bead cannot
        be smaller than the radius. Therefore, clip the data to be at least as
        great as the radius.
        However, values much smaller than the radius could indicate an errornes
        calibration with too small displacement sensitivities, which would lead
        to too small displacements in Z and in turn to negative distances.
        Therefore, if you want to check for this kind of error, switch off the
        clip_Z functionality.
    """
    if focalshift is None:
        focalshift = calibration.focalshift
    if radius is None:
        radius = calibration.radius
    positionZ = positionXYZ[:,2]

    displacementXYZ = calibration.displacement(psdXYZ, positionZ=positionZ)

    # distance, point of attachment of DNA
    # displacement, displacement of bead out of trap center
    # radius
    distanceXYZ = positionXYZ.copy()
    # distance from attachment point to center of bead
    # attachmentXY - displacementXY
    distanceXYZ[:, 0:2] -= displacementXYZ[:, 0:2]
    # Direction of distance vector is from attachment point
    # to the bead center
    distanceXYZ[:, 0:2] *= -1

    # If the bead is free (i.e. above the surface with a distance Z > 0), a
    # movement of the positionZ leads to a distance change reduced by the focal
    # shift.
    # If the bead is on the surface, a movement of the positionZ leads to a
    # distance change independant of the focal shift.
    idx_free = positionXYZ[:, 2] < 0
    idx_touch = positionXYZ[:, 2] >= 0
    distanceXYZ[idx_free, 2] = (- positionXYZ[idx_free, 2] * focalshift
                                + radius
                                # distanceZ + radius + displacementZ
                                + displacementXYZ[idx_free, 2])
    distanceXYZ[idx_touch, 2] = (- positionXYZ[idx_touch, 2]
                                 + radius
                                 # distanceZ + radius + displacementZ
                                 + displacementXYZ[idx_touch, 2])

    if clip_Z:
        distanceXYZ[:, 2] = distanceXYZ[:, 2].clip(min=radius)
    # A positive positionZ signal (positionZ upwards) corresponds to a
    # decreasing (negative) distance of the bead to the surface:
    #   -> distanceZ ~ - positionZ
    # A movement of the positionZ gets reduced by the focalshift:
    #   -> distanceZ = - positionZ * focalshift
    # The distanceZ (positionZ) is 0, where the bead touches the surface:
    #   -> center of the bead is at distanceZ + radius
    # The bead is stressed down with increasing positive distanceZ. This leads
    # to a negative displacement, which reduces the distance:
    #   -> distanceZ + displacement

    # distance from attachment point to bead center
    return distanceXYZ


def displacement(displacementXYZ, positionXY, posmin=10e-9):
    """
    Calculate the displacemnet of the microsphere.

    Parameters
    ----------
    displacementXYZ : 2D numpy.ndarray of type float
        displacementXYZ.shape[1] can consist of either 3 (XYZ) or 2 (XY) axes
    posmin : float
        The `posmin` is used to decide wether the magnitude of the force has to
        be corrected with the sign depending on the position. The `posmin` sets
        the value the position signal has to be deflected at least to be
        counted as active pulling on the bead. The value should at least be >=
        12 times the standard deviation of the unexcited position signal.
        Smaller values could (depending on the number of datapoints) possibly
        lead to falsly detected excitation of the signal.
    """
    return - distance(displacementXYZ, positionXY, posmin=posmin)


def distance(distanceXYZ, positionXY, posmin=10e-9):
    """
    Calculate the distance of the attachment point to the bead center.

    Parameters
    ----------
    distanceXYZ : 2D numpy.ndarray of type float
        distanceXYZ.shape[1] can consist of either 3 (XYZ) or 2 (XY) axes
    posmin : float
        The `posmin` is used to decide wether the magnitude of the force has to
        be corrected with the sign depending on the position. The `posmin` sets
        the value the position signal has to be deflected at least to be
        counted as active pulling on the bead. The value should at least be >=
        12 times the standard deviation of the unexcited position signal.
        Smaller values could (depending on the number of datapoints) possibly
        lead to falsly detected excitation of the signal.
    """
    if posmin is None:
        posmin = 0
    # The sign of the magnitude, i.e. the direction of the distance is
    # important for the noise around +/- 0 N.
    # The sign of the magnitude of the distanceXY depends on the positionXY. If
    # we pull the bead to one side and the bead is displaced to the same side,
    # we get a positive magnitude of the distance. If the bead is displaced to
    # the opposite side of the one we are pulling to, we get a negative
    # magnitude of the distance. Keep in mind that a positive displacement
    # results in an opposite directed pointing distance. The sign of the
    # magnitude of the distanceZ is independent of the positionZ.
    signD = np.sign(distanceXYZ)
    # The position determines the direction of the force only of the axes where
    # we actively pull on the bead
    posmax = np.max(np.abs(positionXY), axis=0)
    idx = posmax >= posmin
    signD[:, XY][:, idx] *= - np.sign(positionXY[:, idx])

    # Square the distances and account for the signs
    distance_sq = distanceXYZ**2
    # Calculate a "weighted" sign. Greater distances have greater influence on
    # the final sign
    dist_sq_sum = np.sum(distance_sq, axis=1)
    sign_dist_sum = np.sign(np.sum(distance_sq * signD, axis=1))
    return np.sqrt(dist_sq_sum) * sign_dist_sum


def extension(distance, radius):
    """
    Calculate the extension of the DNA by simply subtracting the radius from
    the distance.
    """
    return distance - radius


def angle(force, forceXYZ, excited_axis, distance, distanceXYZ):
    """
    Returns a dictionary of angles, for three corners (A), (B), and (C),
    calculated by (F)orce and (D)istance:
        B
        |\
      a | \ c
        |  \
        |___\
       C  b  A
    """

    Fabs = force
    Fz = forceXYZ[:, 2]
    if excited_axis == 'X':
        ea = 0
    else:
        ea = 1
    Fxy = forceXYZ[:, ea]

    dist = distance
    dist3D = distanceXYZ
    Dz = dist3D[:, 2]
    Dxy = dist3D[:, ea]

    AF = _angle(Fz, Fxy, Fabs, angle='A')
    BF = _angle(Fz, Fxy, Fabs, angle='B')
    CF = _angle(Fz, Fxy, Fabs, angle='C')
    AD = _angle(Dz, Dxy, dist, angle='A')
    BD = _angle(Dz, Dxy, dist, angle='B')
    CD = _angle(Dz, Dxy, dist, angle='C')

    angle = dict(list(zip(['AF', 'BF', 'CF', 'AD', 'BD', 'CD'],
                          [AF, BF, CF, AD, BD, CD])))

    return angle


def _angle(a, b, c, angle='A'):
    """
        B
        |\
      a | \ c
        |  \
        |___\
       C  b  A
    """
    if angle == 'A':
        _a = b
        _b = c
        _c = a
    elif angle == 'B':
        _a = a
        _b = c
        _c = b
    else:
        _a = a
        _b = b
        _c = c

    cos_angle = (_a**2 + _b**2 - _c**2) / (2 * _a * _b)
    return np.arccos(np.fabs(cos_angle)) * 180.0 / np.pi



# -*- coding: utf-8 -*-
"""
Created on Wed Mar  9 17:09:09 2016

@author: Tobias Jachowski
"""
import hashlib
import inspect
import numpy as np

from .. import config as cf
from .. import traces as tc
from .region import Region


class Record(Region):
    """
    A Record consists of:
        - a filename, containing the name of the filename
        - a parafile, containing the name of the parafile
        - a calibration object, containing dsurf, focalshift and radius
          (and the height dependent stiffnesses and displacement sensitivities)
        - the data loaded from the filename
        - the parameters loaded from the parafile
        - and some methods for convenient access of the data and para.

    A Record contains some basic modifications, like a static and fixed offset,
    conversion, and inversion factor. The offset, inversion, and conversion
    factors depend on the parameters given by the experimental setup and
    software, only.
    The offset factor should be applied only, if the graphical user interface
    of the OT has a different value than the final value stored in the data.
    The inversion factor should be applied only, if the values e.g. of the
    position are moving in an inverted fashion.
    The conversion factor should be applied only, if the values e.g. of the
    position are stored in Volts. The position values are needed to be in um.

    The modifications are needed to have the same values as used during the
    calibration and need to be set accordingly.

    For future development one should consider to implement the OT control
    software such that whereever possible, Volts are converted to SI units and
    a common convention for the directions (left/right, down/up is neg/pos) is
    followed.
    """
    def __init__(self, datasource, traces, calibration, offset=None,
                 conversion=None, inversion=None, **kwargs):
        """
        Parameters
        ----------
        datasource : DataSource
            The object to retrieve data from. This object has to have the
            function as_array() and the attribue `name`.
        offset: dictionary of trace: value pairs
            value can either be a value or a numpy function (e.g. median)

        """
        super().__init__(max_parents=1, caching=True, **kwargs)

        if datasource is None:
            raise TypeError("Record missing the required positional argument "
                            "'datasource'.")
        self._datasource = datasource
        self._dataident = None

        # Instance of Calibration used to get dsurf, radius and focalshift
        if calibration is None:
            raise TypeError("Record missing the required positional argument "
                            "'calibraion'.")
        self.calibration = calibration

        # Initialize self._shape before trying to set offset, inversion, or
        # conversion, which in turn need to (indirectly) access self._shape.
        # ZODB volatile.
        self._v_data_cached = self._raw_data
        raw_data = self._v_data_cached
        self._shape = raw_data.shape

        # Check if there are as many descriptions as traces in the data
        traces = tc.normalize(traces)
        if len(traces) != self.num_traces:
            raise TypeError('Record argument "traces" has {} elements, but '
                            '"datasource" has {} data columns. Please, '
                            'provide the argument "traces" with a list of '
                            'names for every data column, or edit the '
                            '"cfgfile", accordingly'.format(len(traces),
                                                            self.num_traces))
        self._traces = traces

        # Initialize values of offset (0), inversion (1), and conversion (1)
        # for all available traces
        self._offset = np.zeros(self.num_traces)
        self._conversion = np.ones(self.num_traces)
        self._inversion = np.ones(self.num_traces)

        # Modify offset, conversion and inversion according to optional
        # parameters
        try:
            for trace, value in offset.items():
                if isinstance(value, str):
                    trace_idx = self.traces_to_idx(trace)
                    value = getattr(np, value)(raw_data[:, trace_idx], axis=0)

                self.set_offset(trace, value)
        except:
            pass

        try:
            for trace, value in conversion.items():
                self.set_conversion(trace, value)
        except:
            pass

        try:
            for trace, value in inversion.items():
                self.set_inversion(trace, value)
        except:
            pass

        # reset cache according to modified offset, inversion, and conversion.
        # ZODB volatile.
        self._v_data_cached = ((raw_data - self.offset) * self.inversion
                               * self.conversion)

    def _get_data_uncached(self, samples, traces_idx, copy=True):
        """
        Returns the data of Record, according to self.filename.
        Copy is always True for Record, because every time it reads in the data
        it createas a new numpy.ndarray.
        """
        # read in data
        data = ((self._raw_data[samples, traces_idx] - self.offset[traces_idx])
                * self.inversion[traces_idx]
                * self.conversion[traces_idx])
        # TODO: Implement different samplingrates for different traces. Pandas?
        return data

    @property
    def _raw_data(self):
        """
        Reads and returns uncorrected, uncached data (no offset, no inversion,
        no conversion).
        Additionally, sets self._shape according to _raw_data.shape and creates
        or compares previously generated hashdigest of the data.
        """
        data = self.datasource.as_array()
        if not isinstance(data, np.ndarray):
            raise TypeError("The data you try to load is no numpy array!")
        if data.ndim != 2:
            raise ValueError("The data array you try to load does not have 2 "
                             "dimensions!")
        data = data.copy(order='C')
        ident = hashlib.md5(data).hexdigest()
        if self._dataident is None:
            self._dataident = ident
        elif self._dataident != ident:
            raise ValueError("The data you try to load from '%s' has changed "
                             "since the last time. Please, check the "
                             "datasource of the record '%s'."
                             % (self.datasource.name, self.name))

        return data

    @property
    def offset(self):
        return self._offset.copy()

    @property
    def inversion(self):
        return self._inversion.copy()

    @property
    def conversion(self):
        return self._conversion.copy()

    def set_offset(self, trace, offset):
        """
        Set the offset of a selected trace. The offset should be given in units
        of the raw data, this means independent of inversion and conversion.
        """
        trace = self.traces_to_idx(trace)
        self._offset[trace] = offset
        # Inform descendants of change
        self.set_changed()

    def set_inversion(self, trace, invert=True):
        trace = self.traces_to_idx(trace)
        self._inversion[trace] = - int(invert) or 1
        # Inform descendants of change
        self.set_changed()

    def set_conversion(self, trace, conversion):
        trace = self.traces_to_idx(trace)
        self._conversion[trace] = conversion
        # Inform descendants of change
        self.set_changed()

    @property
    def calibration(self):
        return self.parent

    @calibration.setter
    def calibration(self, calibration):
        if calibration is None:
            raise TypeError("Calibration must not be None.")
        self.set_parent(calibration)

    @property
    def datasource(self):
        return self._datasource

    @property
    def samplingrate(self):
        # TODO: Implement different samplingrates for different traces. Pandas?
        return self.datasource.samplingrate

    @property
    def start(self):
        return 0

    @property
    def stop(self):
        return self._shape[0]

    @property
    def num_traces(self):
        return self._shape[1]

    @property
    def traces(self):
        return self._traces.copy()

    @property
    def caching(self):
        return self._caching

    @caching.setter
    def caching(self, caching):
        super(Record, self.__class__).caching.fset(self, caching)
        if caching is False:
            print("You switched of caching for Record.",
                  "This will seriously lessen the performance of pyoti!",
                  "Do this only, if you have weighty reasons.",
                  "Otherwise, you should revert and reset `caching` to True",
                  sep="\n")


class SpecialRecord(Record):
    """
    This class inherits from the superclass Record and can be used as a
    template to modify the default behaviour of how the data is read in.
    Details see property _raw_data().
    """
    def __init__(self, datasource, traces, calibration, offset=None,
                 conversion=None, inversion=None, **kwargs):

        super().__init__(datasource, traces, calibration, offset=offset,
                         conversion=conversion, inversion=inversion, **kwargs)

        # Get corrected raw_data, which by super().__init__() was initialized
        # and stored in self._v_data_cached
        # raw_data = self._v_data_cached

        # Or get uncorrected raw_data
        # raw_data = self._raw_data()

        # Do something to the data for initialization of SpecialRecord

    @property
    def _raw_data(self):
        data = super()._raw_data
        # do something to data ...
        # e.g. calculate positionXYZ of a combination of signals of one or
        # multiple stage, objective, moveable lens, or mirror signals. Keep in
        # mind that positionZ is increasing for increasing distances and
        # decreasing for decreasing distances of the bead to the surface.
        # PositionXY need to have the same sign as psdXY, i.e. if the bead is
        # pulled to the left/right and displaced to the left/right, positionXY
        # and psdXY both need to have the same signs.
        return data


def create_record(calibration,
                  traces=None,
                  name=None,
                  group=None,
                  rc_class=None,
                  ds_class=None,
                  offset=None,
                  conversion=None,
                  inversion=None,
                  cfgfile=None,
                  **kwargs):
    """
    Parameters
    ----------
    **kwargs
        Is used to get parameters for initialization of `rc_class` and
        initialisation of `ds_class`
    """

    # Set default configfile
    cfgfile = cfgfile or 'record.cfg'

    # Read configfile
    cfg = cf.read_cfg_file(cfgfile)

    print(("Creating record '" + name + "':"))

    rc_class = rc_class or cf.get_cfg_class(cfg, sec='record',
                                            std_mod='.region',
                                            std_cls='Record')
    traces = traces or cf.get_cfg_list(cfg, sec='record', opt='traces')

    if not rc_class:
        print("Could not create Record class defined in config file %s"
              % cfgfile)
        return None

    # get parameters specific for record
    record_pars = {}
    for par in inspect.getargspec(rc_class.__init__)[0]:
        if par in kwargs:
            record_pars[par] = kwargs.pop(par)

    # datasource class and parameters
    ds_class = ds_class or cf.get_cfg_class(cfg, sec='datasource',
                                        std_mod='.plugins.datasources.generic',
                                        std_cls='GenericDataFile')

    if not ds_class:
        print("Could not create DataSource class defined in config file %s"
              % cfgfile)
        return None

    # create datasource
    datasource = ds_class(**kwargs)

    # read and convert offset, conversion, and inversion to dict of floats
    # (booleans)
    offset = offset or cf.get_cfg_sec_dict(cfg, 'offset', convert='float')
    conversion = conversion or cf.get_cfg_sec_dict(cfg, 'conversion',
                                                   convert='float')
    inversion = inversion or cf.get_cfg_sec_dict(cfg, 'inversion',
                                                 convert='boolean')

    record = rc_class(datasource, traces, calibration, offset=offset,
                      conversion=conversion, inversion=inversion, name=name,
                      group=group, **record_pars)

    return record



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 10 21:33:22 2016

@author: Tobias Jachowski
"""
import matplotlib.pyplot as plt
import numpy as np

from pyoti.modification.modification import Modification, GraphicalMod
from pyoti import traces as tc
from pyoti.evaluate import signal as sn


class IAttachment(GraphicalMod):
    """
    Subclass of Attachment that provides graphical interfaces to adjust the fit
    parameters.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Define some transient parameters needed for the graphical fitting of
        # the attachment plateaus [excited_psd, excited_position] for left and
        # right stress views
        self.rightposition = None
        self.leftposition = None
        self.rightpsd = None
        self.leftpsd = None

        # Dimensions of the buttons to adjust the plateaus:
        self.left = 0.13
        self.bottom = 0.79
        self.width = 0.0625
        self.height = 0.046875
        self.bspace = 0.01
        self.lspace = 0.01333

        self._lines = {}
        self._ax = None
        self.supertitle = None
        self._button = {}

        # create some properties of actions/corresponding buttons
        action  = [ 'left',  'right', 'lleft', 'rright', 'up',    'down'  ]
        label   = [ '<',     '>',     '<<',    '>>',     'up',    'down'  ]
        offsetP = [  0.0,     0.0,     0.0,     0.0,     -0.0025,  0.0025 ]
        offsetS = [  2.5e-9, -2.5e-9,  25e-9,  -25e-9,    0.0,     0.0    ]
        row     = [  0,       0,       1,       1,        1,       0      ]
        column  = [  0,       2,       0,       2,        1,       1      ]
        self.action  = action
        self.label   = dict(list(zip(action, label)))
        self.offsetP = dict(list(zip(action, offsetP)))
        self.offsetS = dict(list(zip(action, offsetS)))
        self.row     = dict(list(zip(action, row)))
        self.column  = dict(list(zip(action, column)))

    def _figure(self):
        """
        Initialize and show an interactive plot for adjusting the attachment.
        Adjust the plateau correction parameters interactively and set the
        modification accordingly.
        The plot is stored in self.figure.
        """
        # create new figure and axes for adjusting the plateaus
        figure, ax = plt.subplots(1, 1, sharex=True, sharey=True)
        self._ax = ax

        # create buttons for interactive correction of plateaus and assign
        # correct functions
        # see http://math.andrej.com/2009/04/09/pythons-lambda-is-broken/ for
        # explanation of weird function assignment
        ax_button = {}
        for ac in self.action:
            ax_button[ac] = figure.add_axes([self.column[ac] *
                                             (self.lspace + self.width) +
                                             self.left, self.row[ac] *
                                             (self.bspace + self.height) +
                                             self.bottom,
                                             self.width,
                                             self.height])
            self._button[ac] = plt.Button(ax_button[ac], self.label[ac])

            def ap(event, ac=ac):
                self._adjust_plateaus(ac)
            # connect button to action, accordingly
            self._button[ac].on_clicked(ap)

        # create lines to plot the plateaus
        self._lines['left'] = ax.plot([0], [0], 'r', alpha=0.75)[0]
        self._lines['right'] = ax.plot([0], [0], 'g', alpha=0.75)[0]
        ax.ticklabel_format(useOffset=False)
        ax.grid(True)

        self.supertitle = figure.suptitle("Adjust plateaus to make them "
                                          "overlap")

        return figure

    def _update_fig(self, **kwargs):
        """
        Update the plot
        """
        # recalculate the plateaus with the new offset and scaling values
        self.calculate_plateaus()

        # plot data
        self._lines['left'].set_data(self.leftposition * 1e6, self.leftpsd)
        self._lines['right'].set_data(self.rightposition * 1e6, self.rightpsd)
        excited_psd = self.modification.traces_apply[0]
        excited_position = self.modification.traces_apply[1]
        self._ax.set_xlabel(tc.label(excited_position))
        self._ax.set_ylabel(tc.label(excited_psd))

        # recompute ax.dataLim
        self._ax.relim()
        # update ax.viewLim using new dataLim
        self._ax.autoscale_view()

    def _pre_close_fig(self):
        # Store attachment fit plot for ducumentation
        self.supertitle.set_text('Adjusted  plateaus')
        self._lines.clear()
        self._ax = None
        self._button.clear()
        self.supertitle = None

    def _adjust_plateaus(self, action):
        """
        Adjusts the attachment (offset of excited_position) and the the scaling
        factor to correct for differences of left and right DNA overstretching
        plateaus.
        It is interactively called from the data plot (see below) and updates
        the plot accordingly.
        """
        # change offset and scaling for plateaus
        self.modification.iattributes.offsetPsd += self.offsetP[action]
        self.modification.iattributes.offsetStage += self.offsetS[action]

        self.update_fig()

    def calculate_plateaus(self):
        """
        Calculate the plateaus according to the offsets and the scaling of
        data_based.
        """
        # determine the excited axis of position and psd signal
        ex = self.modification._excited()
        excited_psd = self.modification._NAME['psd'][ex]
        excited_position = self.modification._NAME['position'][ex]
        self.modification.traces_apply = [excited_psd, excited_position]

        # recalculate data for plotting
        # data_based: [excited_psd, excited_position]
        data_based = self.modification._get_data_based(
            traces=self.modification.traces_apply, window=False, decimate=True)

        # subtract offsets
        data_based[:, 0] -= self.modification.iattributes.offsetPsd
        data_based[:, 1] -= self.modification.iattributes.offsetStage

        # determine left/right stress regions of DNA
        signal = data_based[:, 1]  # [excited_position]
        resolution = self.modification.view_based.samplingrate \
            / self.modification.decimate
        minima, maxima = sn.get_extrema(signal, resolution)
        rightstress, _, leftstress = \
            sn.get_sections(signal, minima, maxima)[1][0:3]

        # invert the signals of left stress cycle
        data_based[leftstress] *= -1

        # set plateau data arrays
        self.rightposition = data_based[rightstress, 1]
        self.leftposition = data_based[leftstress, 1]
        self.rightpsd = data_based[rightstress, 0]
        self.leftpsd = data_based[leftstress, 0]


class Attachment(Modification):
    """
    Determine attachment point of DNA and scaling of lateral PSD
    """
    GRAPHICALMOD = IAttachment

    def __init__(self, db_update=False, **kwargs):
        super().__init__(datapoints=25000, **kwargs)

        # determine the excited axis of position and psd signal
        if not db_update:
            ex = self._excited()
            excited_psd = self._NAME['psd'][ex]
            excited_position = self._NAME['position'][ex]
            self.traces_apply = [excited_psd, excited_position]

        # Define parameters that are used to calculate the modification
        # offset of PSD relative to trap center position of bead
        self.add_iattribute('offsetPsd', description='Offset PSD (V)',
                            value=0.0)
        # offset of the position relative to the attachment point of the DNA
        self.add_iattribute('offsetStage', description='Offset position (m)',
                            value=0.0)

    def _print_info(self):
        print("    Excited axis is: %s" % self.traces_apply[1])

    def _modify(self, data, samples, data_traces, data_index, mod_index):
        # correct attachment point of DNA: adjust excited_position
        # (set it to 0 where DNA is attached)
        # correct for offset of excited_psd
        offset = np.array([self.iattributes.offsetPsd,  # excited_psd,
                           self.iattributes.offsetStage])  # excited_position

        data[:, data_index] -= offset[np.newaxis, mod_index]

        return data


# The following is only to update to database version 0.8.0
class GAttachment(Attachment):
    pass



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 10 21:11:44 2016

@author: Tobias Jachowski
"""
import numpy as np

from pyoti.modification.modification import Modification
from pyoti import traces as tc


class Offset(Modification):
    """
    Modification that corrects the offset of psdX, psdY, and psdZ.
    """
    def __init__(self, **kwargs):
        traces_apply = ['psdX', 'psdY', 'psdZ']
        super().__init__(automatic_switch=True, traces_apply=traces_apply,
                         **kwargs)

        # register widgets for the offsets
        for trace in self.traces_apply:
            key = self._key(trace)
            description = ''.join(('Offset ', tc.label(trace)))
            self.add_iattribute(key, description=description, value=0.0)

    def _recalculate(self):
        # Calculate modification for offset from the selected offset span
        traces = self.traces_apply
        data = self._get_data_based(traces=traces, copy=False)
        # should have the same dimension as traces_apply
        self.set_offset(data.mean(axis=0), leave_automatic=True)

    def _modify(self, data, samples, data_traces, data_index, mod_index):
        data[:, data_index] -= self.offset[np.newaxis, mod_index]
        return data

    def _key(self, trace):
        return ''.join(('offset_', trace))

    def get_offset(self):
        return np.array([self.iattributes[self._key(trace)]
                         for trace in self.traces_apply])

    def set_offset(self, offset, trace=None, leave_automatic=False):
        # trace should have the same dimension as traces_apply
        if trace is None:
            traces = self.traces_apply
        else:
            traces = [trace]

        for trace in traces:
            index = self.lia(trace)
            key = self._key(trace)
            self.iattributes.set_value(key, offset[index],
                                       leave_automatic=leave_automatic)

    offset = property(get_offset, set_offset)



# -*- coding: utf-8 -*-
# """
# - Author: steve simmert
# - E-mail: steve.simmert@uni-tuebingen.de
# - Copyright: 2015
# """
"""
Provides functions to find the focal shift of an optical tweezers setup.
"""

from . import ureg

from lmfit import minimize
from lmfit import Parameters
from lmfit import report_fit

import matplotlib.pyplot as plt

from scipy import exp
from scipy import mean
from scipy import pi
from scipy import sin

from .plotting import add_plot_to_figure


def fs_fun(params, heights, wavelength, ref_index, signal=None, simple=False):
    """
    Calculates the difference (residual) between the predicted intensity
    from the model function and the experimental data.

    Arguments
    ---------
    params : lmfit.Parameters
        Object that hold "lmfit.Parameter()"
        Parameters are:
            focal_shift, U, tau, h0, p0, .., p3
    height : array(float)
        Apparent distance between the surface and the bead center in um.
    wavelength : float
        The wavelength of the trapping laser in nm.
    ref_index : float
        Refractive index of the sample medium.
    signal : array(float) or None
        Intensity-signal to calculate the residuals to. If None, the model for
        the given parameters is returned.
    """
    h = heights
    p = params.valuesdict()
    fs = p['focal_shift']
    U = p['U']
    tau = p['tau']
    h0 = p['h0']
    p0 = p['p0']
    if not simple:
        p1 = p['p1']
        p2 = p['p2']
        p3 = p['p3']

    wl = wavelength * 1e-3 / (2 * ref_index)

    if simple:
        model = (U * exp(-(h - h0)**2 / tau) *
                 sin((2 * pi / wl) * fs * (h - h0)) +
                 p0)
    else:
        model = (U * exp(-(h - h0)**2 / tau) *
                 sin((2 * pi / wl) * fs * (h - h0)) +
                 p0 + p1 * h + p2 * h**2 + p3 * h**3)

    if signal is None:
        return model

    return (signal - model)


def gen_fs_fit_pars(focal_shift=0.8,
                    U=0.5,  # Volts
                    h0=0.0,  # stage offset in um
                    tau=1.0,  # decay rate of gaussian slope
                    p0=8.5,  # mean intensity signal
                    p1=0.01,  # polynomial parameters
                    p2=0.01,
                    p3=0.001,
                    simple=False
                    ):
    """ Generate the lmfit fitting Parameters object. """
    pars = Parameters()
    pars.add('focal_shift', value=focal_shift, min=0, max=1.0)
    pars.add('U', value=U)
    pars.add('h0', value=h0)
    pars.add('tau', value=tau, min=0)
    pars.add('p0', value=p0)
    if not simple:
        pars.add('p1', value=p1)
        pars.add('p2', value=p2)
        pars.add('p3', value=p3)

    return pars


def get_focal_shift(heights,  # in um
                    signal,
                    wavelength,  # in nm
                    ref_index,
                    simple=False,
                    height_unit='um',
                    wavelength_unit='nm',
                    parameters=None,
                    report_focal_shift=False,
                    plot_fit=False,
                    plt_axis=None,
                    verbose=False,
                    **init_kwargs
                    ):
    """
    Find the focal shift of the optical trap that is caused due to the
    refravtive index mismatch between coverslip and medium.

    The function uses teh following model to find the focal shift

    Model
        I = (U * exp(-(h - h0)**2 / tau) *
             sin(2 * pi * (h - h0) / wl) +
             p0 + p1 * h + p2 * h**2 + p3 * h**3)
        where wl = wavelength * 1e-3 / (2 * ref_index * fs)

    Arguments
    ---------
    heights : array
        Distances between surface and bead-center.
    signal : array
        Intensities that show interference fluctuations, due to standing
        wave between surface and trapped bead.
    wavelength : float
        Wavelength of the trapping laser.
    ref_index : float
        Refractive index of the medium.
    simple : bool
        If True, the function only uses an offset p0, if False, the function
        uses a 3rd-order polynomial.
    height_unit : str
        Unit of the heights.
    wavelength_unit : str
        Unit of the wavelength.
    parameters : lmfit.Parameters()
        Object with predefined initial guesses of the fit function.
        If None predefined values are used. Use gen_fs_fit_pars() to
        initialize.
    verbose : bool
        Plot fit and give more info if True.

    Keyword Arguments
    -----------------
    focal_shift
    U
    h0
    tau
    p0
    p1
    p2
    p3

    Returns
    -------
    minimizer : lmfit.Minimizer

    See Also
    --------
    gen_fs_fit_pars
    """
    if height_unit != 'um':
        heights *= ureg(height_unit).to('um').magnitude
    if wavelength_unit != 'nm':
        wavelength *= ureg(wavelength_unit).to('nm').magnitude

    # define parameters for the fit
    if parameters is None:
        init_pars = {'focal_shift': 0.8,
                     'U': max(signal)-min(signal),
                     'h0': min(heights)+0.2*(min(heights)+max(heights)),
                     'p0': mean(signal)
                     }
        for k, v in init_pars.items():
            if k not in init_kwargs:
                init_kwargs[k] = v
        pars = gen_fs_fit_pars(simple=simple, **init_kwargs)
    else:
        pars = parameters

    minimizer = minimize(fs_fun,
                         pars,
                         args=(heights,  # in um
                               wavelength,  # in nm
                               ref_index),
                         kws={'signal': signal,
                              'simple': simple})

    minimizer.fcn_eval = fs_fun(minimizer.params,
                                heights,
                                wavelength,
                                ref_index,
                                simple=simple)
    if verbose:
        report_fit(minimizer)
    if plot_fit or verbose:
        if plt_axis:
            fig = plt_axis.figure
        else:
            fig = plt.figure()
        ax = add_plot_to_figure(fig,
                                heights, signal,
                                fmt='o',
                                axis=plt_axis)

        add_plot_to_figure(fig, heights, minimizer.fcn_eval, fmt='-', axis=ax)
        fig.show()
    if minimizer.success:
        fs = minimizer.params['focal_shift'].value
        dfs = minimizer.params['focal_shift'].stderr
        if report_focal_shift or verbose:
            print('Focal shift = {0:1.3f} +/- {1:1.3f}'.format(fs, dfs))
    else:
        if verbose:
            print('Least squares fit did not succeed.')
            print('Minimizer message: "{0:s}"'.format(minimizer.message))
            print('lmdif message: "{0:s}"'.format(minimizer.lmdif_message))

    return minimizer
